import functools
import types


def define_module_exporter():
    all_list = []
    def export(obj):
        all_list.append(obj.__name__)
        return obj
    return export, all_list


def copy_func(fn):
    if type(fn) is not types.FunctionType:
        return functools.wraps(fn)(lambda *args, **kwargs: fn(*args, **kwargs))
    copy = type(lambda: None)(
        fn.__code__,
        fn.__globals__,
        name=fn.__name__,
        argdefs=fn.__defaults__,
        closure=fn.__closure__
    )
    copy = functools.update_wrapper(copy, fn)
    copy.__kwdefaults__ = fn.__kwdefaults__
    return copy


def alias_for(orig_fn):
    def alias_for_orig_func(alias_fn):
        new_fn = functools.wraps(alias_fn)(copy_func(orig_fn))
        docstring_header = alias_fn.__doc__
        if docstring_header is None:
            docstring_header = 'Alias for {orig_qual_name} ({orig_module_name} :: {orig_name}).'.format(
                orig_qual_name=orig_fn.__qualname__,
                orig_module_name=orig_fn.__module__,
                orig_name=orig_fn.__name__
            )
        docstring_footer = orig_fn.__doc__

        if docstring_footer is None:
            new_fn.__doc__ = docstring_header
        else:
            new_fn.__doc__ = '\n'.join((
                docstring_header,
                '',
                '[Original documentation]',
                docstring_footer
            ))
        return new_fn
    return alias_for_orig_func


class Flag(object):
    __slots__ = ('_val',)

    def __init__(self, val=False):
        self._val = val

    def __bool__(self):
        return bool(self._val)

    def set(self):
        self._val = True
        return self

    def clear(self):
        self._val = False
        return self



import itertools
import operator
import collections

from iter2.utils import alias_for, define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup export


# Aliases
builtin_map = map
builtin_sorted = sorted
builtin_zip = zip


itertools_chain_from_iterable = itertools.chain.from_iterable
itertools_product = itertools.product

collections_Counter = collections.Counter
operator_itemgetter = operator.itemgetter


@export_from_module
@alias_for(itertools_product)
def cartesian_product(*iterables, repeat=1):
    '''
    Builds cartesian product from iterables. Alias for itertools::product.

    :param iterables:
    :param repeat:  number of `iterables` repetitions to use
    :return:  iterator through `tuples` from cartesian product of `iterables`
    '''
    pass


@export_from_module
def sort_together(*iterables, key_list=(0,), key_func=None, reverse=False):
    '''
    Sorts iterables as they were zipped.

    :param iterables:
    :param key_list:   indexes of iterators which items are used in `key_func`
    :param key_func:  function to calculate a key
    :param reverse:  descending sort order if True
    :return:  iterator of iterators (count = len(iterables))

    Example:
    >>> tuple(map(tuple, sort_together([1, 2, 3, 4, 5], 'abcde', key_func=(lambda x: x % 2))))
    ((2, 4, 1, 3, 5), ('b', 'd', 'a', 'c', 'e'))
    '''
    if key_func is None:
        key = operator_itemgetter(*key_list)
    else:
        if len(key_list) == 1:
            idx = key_list[0]
            key = lambda item: key_func(item[idx])
        else:
            base_key = operator_itemgetter(*key_list)
            key = lambda item: key_func(*base_key(item))
    return builtin_zip(
        *builtin_sorted(
            builtin_zip(*iterables),
            key=key,
            reverse=reverse
        )
    )


@export_from_module
def unique_to_each(*iterables):
    '''
    Searches items which are contained by one iterable only. Order of items in result can differ from original.

    :param iterables:
    :return:  iterator of tuples with unique-to-each items per iterable

    Example:
    >>> tuple(unique_to_each((1, 2, 3), (3, 4, 5), (5, 6, 7)))
    ((1,), (4,), (7,))
    '''
    unique_by_iterable = tuple(builtin_map(frozenset, iterables))
    counted = collections_Counter(itertools_chain_from_iterable(unique_by_iterable))
    unique_to_each = frozenset(item for item, count in counted.items() if count == 1)
    return (
        tuple(unique_to_each & it)
        for it in unique_by_iterable
    )




import random
import unittest

from iter2.option import None2, Some2
from iter2 import iter2


class TestIterator2(unittest.TestCase):
    # ====================
    # `iterator`-behaviour
    # ====================
    def test_next(self):
        it = iter2([1, 2])
        with self.subTest('next-protocol'):
            self.assertEqual(next(it), 1)
        with self.subTest('next-method'):
            self.assertEqual(it.next(), 2)
        with self.subTest('StopIteration on empty iterator'):
            with self.assertRaises(StopIteration):
                it.next()

    # =================================
    # Processing "sequence -> sequence"
    # =================================
    def test_accumulate(self):
        with self.subTest('default: cumulative sum'):
            self.assertEqual(
                iter2([1, 2, 3]).accumulate().to_tuple(),
                (1, 3, 6)
            )
        with self.subTest('custom: with user defined function'):
            self.assertEqual(
                iter2([2, 1, 3]).accumulate(min).to_tuple(),
                (2, 1, 1)
            )
        with self.subTest('w/ initial element'):
            self.assertEqual(
                iter2([2, 1, 3]).accumulate(min, initial=100500).to_tuple(),
                (100500, 2, 1, 1)
            )

    def test_collapse(self):
        pass

    def test_enumerate(self):
        with self.subTest('default enumerate: from 0'):
            self.assertEqual(
                iter2([1, 2, 3]).enumerate().to_tuple(),
                ((0, 1), (1, 2), (2, 3))
            )
        with self.subTest('general enumerate: from 4'):
            self.assertEqual(
                iter2([1, 2, 3]).enumerate(count_from=4).to_tuple(),
                ((4, 1), (5, 2), (6, 3))
            )

    def test_filter(self):
        self.assertEqual(
            iter2([1, 2, 3]).filter(lambda x: x % 3 == 1).to_tuple(),
            (1,)
        )

    def test_filter_builtin_none(self):
        self.assertEqual(
            iter2([0, 1, None, 3, '', {}, set()]).filter_builtin_none().to_tuple(),
            (0, 1, 3, '', {}, set())
        )

    def test_filter_none(self):
        self.assertEqual(
            iter2([Some2(1), None2, Some2(3)]).filter_none().to_tuple(),
            (1, 3)
        )

    def test_flatmap(self):
        self.assertEqual(
            iter2([1, 2, 3]).flatmap(lambda x: [1] * x).to_tuple(),
            (1,) * (1+2+3)
        )

    def test_flatten(self):
        self.assertEqual(
            iter2([(1, 2), (3,), (4, 5)]).flatten().to_tuple(),
            (1, 2, 3, 4, 5)
        )

    def test_for_each(self):
        res = []
        iter2([1, 2, 3]).for_each(res.append)
        self.assertEqual(res, [1, 2, 3])

    def test_map(self):
        self.assertEqual(
            iter2([1, 2, 3]).map(lambda x: x ** 2).to_tuple(),
            (1, 4, 9)
        )

    def test_starmap(self):
        self.assertEqual(
            iter2([(2, 5), (3, 2), (10, 3)]).starmap(pow).to_tuple(),
            (32, 9, 1000)
        )

    # ===================
    # Checking conditions
    # ===================
    def test_all(self):
        self.assertTrue(
            iter2([1, 2, 3]).all(lambda x: x > 0)
        )

    def test_any(self):
        self.assertTrue(
            iter2([1, 2, 3]).any(lambda x: x > 1)
        )

    def test_none(self):
        self.assertTrue(
            iter2([1, 2, 3]).none(lambda x: x < 0)
        )

    # =============
    # Making groups
    # =============
    def test_batching(self):
        'Not implemented'
        pass

    def test_chunks(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3, 4, 5])
                    .chunks(2)
                    .to_tuple(),
                ((1, 2), (3, 4))
            )
        with self.subTest('allow partial'):
            self.assertEqual(
                iter2([1, 2, 3, 4, 5])
                    .chunks(2, allow_partial=True)
                    .to_tuple(),
                ((1, 2), (3, 4), (5,))
            )

    def test_consecutive_groups(self):
        # TODO: implement
        pass

    def test_dedup(self):
        self.assertEqual(
            iter2([1, 2, 2, 3, 3, 3, 2, 2, 1]).dedup().to_tuple(),
            (1, 2, 3, 2, 1)
        )

    def test_group_by(self):
        with self.subTest('default: key is identity function'):
            self.assertEqual(
                iter2([1, 2, 2, 3, 3, 3])
                    .group_by()
                    .starmap(lambda val, subiter: (val, subiter.count()))
                    .to_tuple(),
                ((1, 1), (2, 2), (3, 3))
            )
        with self.subTest('w/ key function'):
            self.assertEqual(
                iter2([1, 2, 2, 3, 3, 3])
                    .group_by(key=lambda x: x % 2)
                    .starmap(lambda val, sub_iter: (val, sub_iter.count()))
                    .to_tuple(),
                ((1, 1), (0, 2), (1, 3))
            )

    def test_group_by_and_transform(self):
        self.assertEqual(
            iter2('AaaABbBCc')
                .group_by_and_transform(keyfunc=str.lower, valuefunc=str.upper)
                .starmap(lambda k, v: (k, v.join('')))
                .to_tuple(),
            (('a', 'AAAA'), ('b', 'BBB'), ('c', 'CC'))
        )

    def test_grouper(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]).grouper(2).to_tuple(),
                ((1, 2), (3, None))
            )
        with self.subTest('w/ fillvalue'):
            self.assertEqual(
                iter2([1, 2, 3]).grouper(2, fillvalue=100500).to_tuple(),
                ((1, 2), (3, 100500))
            )
        with self.subTest('w/o remainder'):
            self.assertEqual(
                iter2([1, 2, 3, 4]).grouper(2).to_tuple(),
                ((1, 2), (3, 4))
            )

    # ===========
    # (Un)Merging
    # ===========
    def test_distribute(self):
        first, second, third = iter2([1, 2, 3, 4, 5]).distribute(3)
        self.assertEqual( first.to_tuple(), (1, 4))
        self.assertEqual(second.to_tuple(), (2, 5))
        self.assertEqual( third.to_tuple(), (3,))

    def test_cartesian_product(self):
        self.assertEqual(
            iter2([1, 2]).cartesian_product(repeat=2).to_tuple(),
            ((1, 1), (1, 2), (2, 1), (2, 2))
        )

    def test_interleave(self):
        seq1 = 'abc'
        seq2 = 'ij'
        seq3 = 'xyz'
        self.assertEqual(
            iter2(seq1).interleave(seq2, seq3).join(''),
            'aixbjycz'
        )

    def test_interleave_shortest(self):
        seq1 = 'abc'
        seq2 = 'ij'
        seq3 = 'xyz'
        self.assertEqual(
            iter2(seq1).interleave_shortest(seq2, seq3).join(''),
            'aixbjy'
        )

    def test_intersperse(self):
        with self.subTest('empty iterator'):
            self.assertEqual(
                iter2(()).intersperse(0).to_tuple(),
                ()
            )
        with self.subTest('common case'):
            self.assertEqual(
                iter2([1, 2, 3]).intersperse(0).to_tuple(),
                (1, 0, 2, 0, 3)
            )

    def test_pairwise(self):
        self.assertEqual(
            iter2([1, 2, 3]).pairwise().to_tuple(),
            ((1, 2), (2, 3))
        )

    def test_partition(self):
        t, f = iter2([1, 2, 3, 4, 5]).partition(lambda x: x % 2 == 0)
        self.assertEqual(t.to_tuple(), (2, 4))
        self.assertEqual(f.to_tuple(), (1, 3, 5))

    def test_permutations(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]).permutations().to_tuple(),
                ((1, 2, 3), (1, 3, 2),
                 (2, 1, 3), (2, 3, 1),
                 (3, 1, 2), (3, 2, 1))
            )
        with self.subTest('pairs'):
            self.assertEqual(
                iter2([1, 2, 3]).permutations(repeat=2).to_tuple(),
                ((1, 2), (1, 3),
                 (2, 1), (2, 3),
                 (3, 1), (3, 2))
            )

    def test_roundrobin(self):
        # TODO: implement
        pass

    def test_stagger(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]).stagger().to_tuple(),
                ((None, 1, 2), (1, 2, 3))
            )
        with self.subTest('offsets'):
            self.assertEqual(
                iter2.range(8).stagger(offsets=(0, 2, 4)).to_tuple(),
                ((0, 2, 4), (1, 3, 5), (2, 4, 6), (3, 5, 7))
            )
        with self.subTest('longest'):
            self.assertEqual(
                iter2([1, 2, 3]).stagger(longest=True).to_tuple(),
                ((None, 1, 2), (1, 2, 3), (2, 3, None), (3, None, None))
            )

    def test_unzip(self):
        letters = 'abc'
        digits = [1, 2, 3]

        with self.subTest('default: arity=2'):
            ls, ds = iter2(zip(letters, digits)).unzip()
            self.assertEqual(ls.to_tuple(), tuple(letters))
            self.assertEqual(ds.to_tuple(), tuple(digits))

        with self.subTest('arity=3'):
            ls, ds, lsr = iter2(zip(letters, digits,reversed(letters))) \
                .unzip(arity=3)
            self.assertEqual(ls.to_tuple(), tuple(letters))
            self.assertEqual(ds.to_tuple(), tuple(digits))
            self.assertEqual(lsr.to_tuple(), tuple(reversed(letters)))

    def test_zip(self):
        self.assertEqual(
            iter2('abc').zip([1, 2, 3]).to_tuple(),
            (('a', 1), ('b', 2), ('c', 3))
        )

    def test_zip_longest(self):
        self.assertEqual(
            iter2('ab').zip_longest([1, 2, 3]).to_tuple(),
            (('a', 1), ('b', 2), (None, 3))
        )

    def test_zip_offset(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2('0123').zip_offset('abcdef', offsets=(0, 1)).to_tuple(),
                (('0', 'b'), ('1', 'c'), ('2', 'd'), ('3', 'e'))
            )
        with self.subTest('negative offsets'):
            self.assertEqual(
                iter2('0123')
                    .zip_offset('abcdef', offsets=(-1, 0), fillvalue='@')
                    .to_tuple(),
                (('@', 'a'), ('0', 'b'), ('1', 'c'), ('2', 'd'), ('3', 'e'))
            )
        with self.subTest('longest'):
            self.assertEqual(
                iter2('0123')
                    .zip_offset('abcdef', offsets=(-1, 0), fillvalue='@', longest=True)
                    .to_tuple(),
                (('@', 'a'), ('0', 'b'), ('1', 'c'), ('2', 'd'), ('3', 'e'), ('@', 'f'))
            )

    # =================
    # Flow manipulation
    # =================
    def test_adjacent(self):
        pass

    def test_chain(self):
        with self.subTest('chain two rich iterators'):
            self.assertEqual(
                iter2([1, 2, 3]).chain(iter2([4, 5])).to_tuple(),
                (1, 2, 3, 4, 5)
            )
        with self.subTest('chain with iterable'):
            self.assertEqual(
                iter2([1, 2, 3]).chain([4, 5]).to_tuple(),
                (1, 2, 3, 4, 5)
            )

    def test_concat(self):
        with self.subTest('concat two rich iterators'):
            self.assertEqual(
                iter2([1, 2, 3]).concat(iter2([4, 5])).to_tuple(),
                (1, 2, 3, 4, 5)
            )
        with self.subTest('concat with iterable'):
            self.assertEqual(
                iter2([1, 2, 3]).concat([4, 5]).to_tuple(),
                (1, 2, 3, 4, 5)
            )

    def test_consume(self):
        self.assertEqual(
            iter2([1, 2, 3]).consume().to_tuple(),
            ()
        )

    def test_drop(self):
        self.assertEqual(
            iter2([1, 2, 3, 4, 5]).drop(2).to_tuple(),
            (3, 4, 5)
        )

    def test_drop_while(self):
        it = iter2([1, 2, 3, 4, 5])
        self.assertEqual(
            it.drop_while(lambda x: x < 3).to_tuple(),
            (3, 4, 5)
        )

    def test_find(self):
        with self.subTest('item exist'):
            self.assertEqual(
                iter2([1, 2, 3]).find(lambda x: x % 2 == 0).unwrap(),
                2
            )
        with self.subTest('not found'):
            self.assertTrue(
                iter2([1, 2, 3]).find(lambda x: x == 5).is_none()
            )

    def test_first(self):
        self.assertEqual(iter2([1, 2, 3]).first().unwrap(), 1)

    def test_last(self):
        self.assertEqual(iter2([1, 2, 3]).last().unwrap(), 3)

    def test_nth(self):
        it = iter2([1, 2, 3])
        with self.subTest('Valid indexes'):
            self.assertEqual(it.nth(1), 2)
            self.assertEqual(it.nth(0), 3)
        with self.subTest('Invalid index'):
            with self.assertRaises(IndexError):
                _ = it.nth(0)

    def test_position(self):
        # TODO: implement
        pass

    def test_prepend(self):
        # TODO: implement
        pass

    def test_cycle(self):
        with self.subTest('repeat forever in cycle'):
            count = random.randint(1, 5)
            self.assertEqual(
                iter2([1]).cycle().take(count).to_tuple(),
                (1,) * count
            )
        with self.subTest('repeat several times'):
            self.assertEqual(
                iter2([1, 2]).cycle(number=4).to_tuple(),
                (1, 2) * 4
            )

    def test_skip(self):
        'alias for `Iterator2.drop`'
        pass

    def test_skip_while(self):
        'alias for `Iterator2.drop_while`'
        pass

    def test_slice(self):
        with self.subTest('Short variant: only `stop` argument.'):
            self.assertEqual(
                iter2([1, 2, 3]).slice(2).to_tuple(),
                (1, 2)
            )
        with self.subTest('Interval variant: `start`, `stop` without `step`.'):
            self.assertEqual(
                iter2([1, 2, 3, 4, 5]).slice(1, 3+1).to_tuple(),
                (2, 3, 4)
            )
        with self.subTest('Full variant: `start`, `stop` and `step`.'):
            self.assertEqual(
                iter2([1, 2, 3, 4, 5]).slice(1, 3+1, 2).to_tuple(),
                (2, 4)
            )

    def test_sliding_window(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]) \
                    .sliding_window() \
                    .to_tuple(),
                ((1,), (2,), (3,))
            )
        with self.subTest('split into pairs'):
            self.assertEqual(
                iter2([1, 2, 3, 4, 5])
                    .sliding_window(step=2, size=2)
                    .to_tuple(),
                ((1, 2), (3, 4))
            )
        with self.subTest('3-size window'):
            self.assertEqual(
                iter2([1, 2, 3, 4, 5])
                    .sliding_window(size=3)
                    .to_tuple(),
                ((1, 2, 3), (2, 3, 4), (3, 4, 5))
            )
        with self.subTest('step > size is OK'):
            self.assertEqual(
                iter2([1, 2, 3, 4, 5])
                    .sliding_window(step=3, size=2)
                    .to_tuple(),
                ((1, 2), (4, 5))
            )
        with self.subTest('partial window at the end'):
            self.assertEqual(
                iter2([1, 2, 3, 4, 5])
                    .sliding_window(step=2, size=2, allow_partial=True)
                    .to_tuple(),
                ((1, 2), (3, 4), (5,))
            )

    def test_step(self):
        self.assertEqual(
            iter2([1, 2, 3, 4, 5]).step(2).to_tuple(),
            (1, 3, 5)
        )

    def test_sort(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([3, 1, 2]).sort().to_tuple(),
                (1, 2, 3)
            )
        with self.subTest('reverse'):
            self.assertEqual(
                iter2([3, 1, 2]).sort(reverse=True).to_tuple(),
                (3, 2, 1)
            )
        with self.subTest('w/ key'):
            self.assertEqual(
                iter2(dict(a=1, b=2, c=0).items())
                    .sort(key=lambda pair: pair[1])
                    .to_tuple(),
                (('c', 0), ('a', 1), ('b', 2))
            )

    def test_take(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]).take().to_tuple(),
                (1,)
            )
        with self.subTest('take pair'):
            self.assertEqual(
                iter2([2, 3]).take(2).to_tuple(),
                (2, 3)
            )
        with self.subTest('consuming order DOES matter'):
            it = iter2([4, 5])
            sub_it_4 = it.take()  # expected to take `4`
            sub_it_5 = it.take()  # expected to take `5`
            self.assertEqual(
                sub_it_5.to_tuple(),  # really consumed (`4`)
                (4,)
            )
            self.assertEqual(
                sub_it_4.to_tuple(),  # really consumed (`5`)
                (5,)
            )
        with self.subTest('result may be insufficient'):
            self.assertEqual(
                iter2([1, 2, 3]).take(10).to_tuple(),
                (1, 2, 3)
            )

    def test_take_now(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]).take_now().collect(),
                (1,)
            )
        with self.subTest('take pair'):
            self.assertEqual(
                iter2([2, 3]).take_now(2).collect(),
                (2, 3)
            )
        with self.subTest('consuming order DOES NOT matter'):
            it = iter2([4, 5])
            sub_it_4 = it.take_now()  # expected to take `4`
            sub_it_5 = it.take_now()  # expected to take `5`
            self.assertEqual(
                sub_it_5.to_tuple(),  # really consumed (`5`)
                (5,)
            )
            self.assertEqual(
                sub_it_4.to_tuple(),  # really consumed (`4`)
                (4,)
            )
        with self.subTest('result may be insufficient'):
            self.assertEqual(
                iter2([1, 2, 3]).take_now(10).collect(),
                (1, 2, 3)
            )

    def test_take_into_option(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]).take_into_option().unwrap().to_tuple(),
                (1,)
            )
        with self.subTest('take pair'):
            self.assertEqual(
                iter2([1, 2, 3]).take_into_option(2).unwrap().to_tuple(),
                (1, 2)
            )
        with self.subTest('consuming order does not matter'):
            it = iter2([1, 2, 3, 4])
            sub_it_1_2 = it.take_into_option(2)
            sub_it_3_4 = it.take_into_option(2)
            self.assertEqual(
                sub_it_3_4.unwrap().to_tuple(),
                (3, 4)
            )
            self.assertEqual(
                sub_it_1_2.unwrap().to_tuple(),
                (1, 2)
            )
        with self.subTest('None2 if insufficient'):
            self.assertTrue(
                iter2([1, 2]).take_into_option(3).is_none()
            )
        with self.subTest('if insufficient then return back taken items to iterator'):
            it = iter2([1, 2])
            self.assertTrue(it.take_into_option(3).is_none())
            self.assertEqual(
                it.take_into_option(2).unwrap().to_tuple(),
                (1, 2)
            )

    def test_take_last(self):
        with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]).take_last().to_tuple(),
                (3,)
            )
        with self.subTest('take pair'):
            self.assertEqual(
                iter2([1, 2, 3]).take_last(2).to_tuple(),
                (2, 3)
            )

    def test_take_last_into_option(self):
         with self.subTest('default'):
            self.assertEqual(
                iter2([1, 2, 3]).take_last_into_option().unwrap().to_tuple(),
                (3,)
            )
         with self.subTest('take pair'):
            self.assertEqual(
                iter2([1, 2, 3]).take_last_into_option(2).unwrap().to_tuple(),
                (2, 3)
            )
         with self.subTest('if insufficient then return back taken items to iterator'):
            it = iter2([1, 2])
            self.assertTrue(it.take_last_into_option(3).is_none())
            self.assertEqual(
                it.take_last_into_option(2).unwrap().to_tuple(),
                (1, 2)
            )

    def test_take_while(self):
        with self.subTest('default: do not keep border element'):
            it = iter2([1, 2, 3, 4, 5])
            self.assertEqual(
                it.take_while(lambda x: x < 3).to_tuple(),
                (1, 2)
            )
            # `3` was consumed and not placed back
            self.assertEqual(
                it.to_tuple(),
                (4, 5)
            )
        with self.subTest('keep border element'):
            it = iter2([1, 2, 3, 4, 5])
            self.assertEqual(
                it.take_while(lambda x: x < 3, keep_border_element=True).to_tuple(),
                (1, 2)
            )
            # `3` was consumed, but placed back
            self.assertEqual(
                it.to_tuple(),
                (3, 4, 5)
            )

    def test_unique(self):
        self.assertEqual(
            iter2([1, 3, 2, 3, 2, 3]).unique().collect(frozenset),
            frozenset([1, 2, 3])
        )

    # =============
    # Materializing
    # =============
    def test_collect(self):
        with self.subTest('Collecting to default collection (tuple)'):
            self.assertEqual(
                iter2([1, 2, 3]).collect(),
                (1, 2, 3)
            )

        SUBTESTS = (
            (tuple,     [1, 2, 3],              (1, 2, 3)),
            (list,      [1, 2, 3],              [1, 2, 3]),
            (frozenset, [1, 1, 3],              frozenset([1, 3])),
            (set,       [1, 1, 3],              {1, 3}),
            (dict,      [('a', 1), ('b', 2)],   dict(a=1, b=2))
        )

        for factory, input, output in SUBTESTS:
            with self.subTest('Collecting to {factory}'.format(factory=factory)):
                self.assertEqual(
                    iter2(input).collect(factory),
                    output
                )

    def test_collect_recursive(self):
        with self.subTest('default: alias for `collect`'):
            self.assertEqual(
                iter2([1, 2, 3]).collect_recursive(tuple),
                iter2([1, 2, 3]).to_tuple(),
            )

        with self.subTest('recursive: 2 levels'):
            self.assertEqual(
                iter2([
                    iter2([1, 2]),
                    iter2([3, 4]),
                    iter2([5, 6]),
                ]).collect_recursive(tuple, levels=2),
                ((1, 2), (3, 4), (5, 6))
            )

    def test_make_sized(self):
        self.assertEqual(
            iter2([1, 2, 3]).make_sized(list).collect(list),
            [1, 2, 3]
        )

    # ===========
    # Duplication
    # ===========
    def test_copy(self):
        orig = iter2([1, 2, 3])
        copy = orig.copy()
        self.assertEqual(
            orig.to_tuple(),
            (1, 2, 3)
        )
        self.assertEqual(
            copy.collect(list),
            [1, 2, 3]
        )

    def test_tee(self):
        copy1, copy2 = iter2([1, 2, 3]).tee()
        self.assertEqual(
            copy1.to_tuple(),
            (1, 2, 3)
        )
        self.assertEqual(
            copy2.collect(list),
            [1, 2, 3]
        )


    # =======
    # Folding
    # =======
    def test_count(self):
        with self.subTest('empty iterator'):
            self.assertEqual(iter2([]).count(), 0)
        with self.subTest('non-empty iterator'):
            self.assertEqual(iter2([1, 2, 3]).count(), 3)

    def test_fold(self):
        with self.subTest('w/o initial'):
            self.assertEqual(
                iter2([1, 2, 3]).fold(lambda st, x: st + x),
                sum([1, 2, 3])
            )
        with self.subTest('w/ initial'):
            self.assertEqual(
                iter2([1, 2, 3]).fold(lambda st, x: st + x, initial=-3),
                sum([-3,  1, 2, 3])
            )

    def test_join(self):
        with self.subTest('str'):
            self.assertEqual(
                iter2('abc').join('-'),
                'a-b-c'
            )
        with self.subTest('bytes'):
            self.assertEqual(
                iter2([b'a', b'b', b'c']).join(b'-'),
                b'a-b-c'
            )
        with self.subTest('incompatible'):
            with self.assertRaises(TypeError):
                _ = iter2('abc').join(b'-')

    def test_max(self):
        with self.subTest('empty'):
            self.assertEqual(
                iter2([]).max(),
                None
            )
        with self.subTest('empty with default'):
            self.assertEqual(
                iter2([]).max(default=100500),
                100500
            )
        with self.subTest('regular, w/o key'):
            self.assertEqual(
                iter2([1, 2, 3]).max(),
                3
            )
        with self.subTest('regular, w/ key'):
            self.assertEqual(
                # [(1, -1), (0, -2), (1, -3), (0, -4)]
                iter2([1, 2, 3, 4]).max(key=lambda x: (x % 2, -x)),
                1
            )

    def test_min(self):
        with self.subTest('empty'):
            self.assertEqual(
                iter2([]).min(),
                None
            )
        with self.subTest('empty with default'):
            self.assertEqual(
                iter2([]).min(default=100500),
                100500
            )
        with self.subTest('regular, w/o key'):
            self.assertEqual(
                iter2([1, 2, 3]).min(),
                1
            )
        with self.subTest('regular, w/ key'):
            self.assertEqual(
                iter2([1, 2, 3, 4]).min(key=lambda x: (x % 2, -x)),
                # [(1, -1), (0, -2), (1, -3), (0, -4)]
                4
            )

    def test_minmax(self):
        with self.subTest('empty'):
            self.assertEqual(
                iter2([]).minmax(),
                (None, None)
            )
        with self.subTest('empty with default'):
            self.assertEqual(
                iter2([]).minmax(default=100500),
                (100500, 100500)
            )
        with self.subTest('regular, w/o key'):
            self.assertEqual(
                iter2([1, 2, 3]).minmax(),
                (1, 3)
            )
        with self.subTest('regular, w/ key'):
            self.assertEqual(
                iter2([1, 2, 3, 4]).minmax(key=lambda x: (x % 2, -x)),
                # [(1, -1), (0, -2), (1, -3), (0, -4)]
                (4, 1)
            )

    def test_product(self):
        self.assertEqual(
            iter2([1, 2, 3]).product(),
            (1 * 2 * 3)
        )

    def test_reduce(self):
        'alias for `iterator2.fold`'
        pass

    def test_sum(self):
        self.assertEqual(
            iter2([1, 2, 3]).sum(),
            sum([1, 2, 3])
        )

    # ======================
    # Lookahead and lookback
    # ======================
    def test_spy(self):
        with self.subTest('default :: sufficient'):
            it = iter2([1, 2, 3])
            self.assertEqual(
                it.spy(2).unwrap().to_tuple(),
                (1, 2)
            )
            self.assertEqual(it.to_tuple(), (1, 2, 3))
        with self.subTest('default :: insufficient'):
            it = iter2([1, 2, 3])
            self.assertTrue(it.spy(4).is_none())
            self.assertEqual(it.to_tuple(), (1, 2, 3))
        with self.subTest('partial :: sufficient'):
            it = iter2([1, 2, 3])
            self.assertEqual(it.spy(2, allow_partial=True).to_tuple(), (1, 2))
            self.assertEqual(it.to_tuple(), (1, 2, 3))
        with self.subTest('partial :: insufficient'):
            it = iter2([1, 2, 3])
            self.assertEqual(it.spy(4, allow_partial=True).to_tuple(), (1, 2, 3))
            self.assertEqual(it.to_tuple(), (1, 2, 3))



import unittest

import iter2.algorithms.extra as algo


class TestExtraAlgorithms(unittest.TestCase):
    def assertEqualAsTuples(self, first, second):
        self.assertEqual(tuple(first), tuple(second))

    def test_cartesian_product(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.cartesian_product((1, 2), 'ab', [True]),
                ((1, 'a', True),
                 (1, 'b', True),
                 (2, 'a', True),
                 (2, 'b', True))
            )
        with self.subTest('w/ repeat'):
            self.assertEqualAsTuples(
                algo.cartesian_product(range(2), repeat=2),
                ((0, 0), (0, 1), (1, 0), (1, 1))
            )

    def test_sort_together(self):
        with self.subTest('simple'):
            first, second = algo.sort_together(
                [3, 2, 1],
                  'cba'
            )
            self.assertEqualAsTuples(  first,       (1, 2, 3) )
            self.assertEqualAsTuples( second, ('a', 'b', 'c') )

        with self.subTest('by one iterable but w/ key_func'):
            first, second = algo.sort_together(
                [1, 2, 3, 4, 5],
                    'abcde',
                key_func=(lambda x: x % 2)
            )
            self.assertEqualAsTuples(  first, ( 2,   4,   1,   3,   5)  )
            self.assertEqualAsTuples( second, ('b', 'd', 'a', 'c', 'e') )

        with self.subTest('w/ parameters'):
            WOMEN_PENSION_AGE = 60
            names, ages, salaries = algo.sort_together(
                [ 'Lera', 'Nastya', 'Polina'],
                [     25,       22,       24],
                [ 150000,   200000,    40000],
                key_list=(1, 2),
                key_func=lambda age, salary: (WOMEN_PENSION_AGE - age) * salary,
                reverse=True
            )
            self.assertEqualAsTuples( names,    ( 'Nastya', 'Lera', 'Polina' ) )
            self.assertEqualAsTuples( ages,     (       22,     25,       24 ) )
            self.assertEqualAsTuples( salaries, (   200000, 150000,    40000 ) )

    def test_unique_to_each(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                map(sorted, algo.unique_to_each(
                    'abc',   'bcd',    'def'
                )), (
                    ['a'],    [],    ['e','f']
                )
            )


import itertools
import collections
import contextlib
import operator

from . import sequence_to_groups

from iter2.utils import alias_for, define_module_exporter


export_from_module, __all__ = define_module_exporter()


# Aliases
builtin_enumerate = enumerate
builtin_filter = filter
builtin_map = map
builtin_slice = slice
builtin_zip = zip

itertools_accumulate = itertools.accumulate
itertools_chain = itertools.chain
itertools_chain_from_iterable = itertools.chain.from_iterable
itertools_cycle = itertools.cycle
itertools_dropwhile = itertools.dropwhile
itertools_filterfalse = itertools.filterfalse
itertools_groupby = itertools.groupby
itertools_islice = itertools.islice
itertools_starmap = itertools.starmap
itertools_takewhile = itertools.takewhile
itertools_tee = itertools.tee

collections_deque = collections.deque
contextlib_suppress = contextlib.suppress

s2g_chunks = sequence_to_groups.chunks


@export_from_module
def accumulate(iterable, func=None, *, initial=None):
    '''
    Return series of accumulated results of binary `func`.
    If `initial` is present then it will be used as first accumulated value.

    :param iterable:
    :param func:
    :param initial:
    :return:

    Example:
    >>> tuple(accumulate([1, 2, 3]))
    (1, 3, 6)
    >>> tuple(accumulate([1, 2, 3], lambda x, y: x * y))
    (1, 2, 6)
    '''
    if initial is not None:
        it = itertools_chain((initial,), iterable)
    else:
        it = iterable
    if func is None:
        return itertools_accumulate(it)
    else:
        return itertools_accumulate(it, func)


@export_from_module
def add_side_effect(iterable, func, *, chunk_size=None, before=None, after=None):
    '''
    Returns new iterator that calls `func` w/ every item from `iterable` as first argument but yields original items.
    If `chunk_size` is set to value bigger than 1 then calls `func` w/ every chunk of `chunk_size` as first argument.
    If `before` is set then calls `before` in the beginning of iteration.
    If `after` is set then calls `after` in the end of iteration.

    :param iterable:
    :param func:
    :param chunk_size:
    :param before:
    :param after:
    :return:

    Example:
    >>> tuple(add_side_effect(['Julia', 'Angelina', 'Jessica'], lambda w: print('  %s' % w), before=lambda: 'My list:', after=print))
    My list:
      Julia
      Angelina
      Jessica

    ('Julia', 'Angelina', 'Jessica')
    >>> tuple(add_side_effect(range(10), lambda ch: print('Processed: %s' % repr(ch)), chunk_size=4))
    Processed: (0, 1, 2, 3)
    Processed: (4, 5, 6, 7)
    Processed: (8, 9)
    (0, 1, 2, 3, 4, 5, 6, 7, 8, 9)
    '''
    try:
        if before is not None:
            before()
        if chunk_size is None or chunk_size == 1:
            for item in iterable:
                func(item)
                yield item
        else:
            for chunk in s2g_chunks(iterable, chunk_size, allow_partial=True):
                func(chunk)
                yield from chunk
    finally:
        if after is not None:
            after()


@export_from_module
def consume(iterable, number=None):
    '''
    Consumes `number` of items from `iterable`. If `number` is not present then whole iterable is consumed.

    :param iterable:
    :param number:
    :return:

    Example:
    >>> tuple(consume(range(10)))
    ()
    >>> tuple(consume(range(10), 8))
    (8, 9)
    '''
    iterable = iter(iterable)  # it can look like iterable but it's not (e.g. range(10))
    if number is None:
        collections_deque(iterable, maxlen=0)  # fastest full consume ever
    else:
        next(itertools_islice(iterable, number, number), None)
    return iterable


@export_from_module
def cycle(iterable, *, number=None):
    '''
    Returns iterator that yields items from `iterable` in cycle `number` times.
    If `number` is not present then cycle is infinite.

    :param iterable:
    :param number:
    :return:

    Example:
    >>> tuple(cycle([0, 1], number=2))
    (0, 1, 0, 1)
    '''
    if number is None:
        return itertools_cycle(iterable)
    else:
        # TODO: think about (micro)optimization
        return itertools_chain_from_iterable(itertools_tee(iterable, number))


@export_from_module
def dedup(iterable):
    '''
    Reduces consecutive repetitions of items in `iterable` to one. Doesn't make items unique in sequence.

    :param iterable:
    :return:

    Example:
    >>> ''.join(dedup('011011110011'))
    '010101'
    '''
    for val, sub_iter in itertools_groupby(iterable):
        yield val
        consume(sub_iter)


@export_from_module
def difference(iterable, func=operator.sub, *, initial=None):
    '''
    By default, compute the first difference of `iterable` using `operator.sub`. This is the opposite of `accumulate`’s default behavior.
    By default `func` is operator.sub(), but other functions can be specified. They will be applied as follows:
        A, B, C, D, ... --> A, func(B, A), func(C, B), func(D, C), ...
    If `initial` is present then it will be used as first value.

    :param iterable:
    :param func:
    :param initial:
    :return:

    Example:
    >>> tuple(difference([1, 3, 6]))
    (1, 2, 3)

    '''
    if initial is not None:
        it = itertools_chain((initial,), iterable)
    else:
        it = iterable
    a, b = itertools_tee(it)
    try:
        item = next(b)
        return itertools_chain((item,), builtin_map(lambda x: func(x[1], x[0]), builtin_zip(a, b)))
    except StopIteration:
        return iter(())


@export_from_module
def drop(iterator, number=1):
    '''
    Skips `number` of items from `iterator`.

    :param iterator:
    :param number:
    :return:

    Example:
    >>> tuple(drop(range(10), 8))
    (8, 9)
    '''
    return itertools_islice(iterator, number, None)


@export_from_module
def drop_while(iterator, predicate):
    '''
    Skips items from `iterator` while `predicate` is true.

    :param iterator:
    :param predicate:
    :return:

    Example:
    >>> tuple(drop_while('aaaaAAAA'), str.islower)
    ('A', 'A', 'A', 'A')
    '''
    return itertools_dropwhile(predicate, iterator)


@export_from_module
def enumerate(iterable, *, count_from=0):
    '''
    Returns iterator that yields pairs (index, item) where index starts from `count_from`.

    :param iterable:
    :param count_from:
    :return:

    Example:
    >>> tuple(enumerate('ab'))
    ((0, 'a'), (1, 'b'))
    '''
    return builtin_enumerate(iterable, count_from)


@export_from_module
def filter(iterable, predicate, *, inverse=False):
    '''
    Returns iterator that yields `predicate`-positive items from `iterable`.
    If `inverse` is true then yields `predicate`-negative items.

    :param iterable:
    :param predicate:
    :param inverse:
    :return:

    Example:
    >>> tuple(filter('aAaA', str.isupper))
    ('A', 'A')
    >>> tuple(filter('aAaA', str.isupper, inverse=True))
    ('a', 'a')
    '''
    if inverse:
        func = itertools_filterfalse
    else:
        func = builtin_filter
    return func(predicate, iterable)


@export_from_module
def filter_none(iterable):
    '''
    Filters builtin `None` values from `iterable`.

    :param iterable:
    :return:

    Example:
    >>> tuple(filter_none([1, None, 2, None]))
    (1, 2)
    '''
    return (
        val
        for val in iterable
        if val is not None
    )


@export_from_module
def flatmap(iterable, func):
    '''
    Assuming `func` produces iterables, applies `func` to each item in `iterable` and yields new items one-by-one from new iterables.
    Is equivalent to flatten(map(iterable, func)).

    :param iterable:
    :param func:
    :return:

    Example:
    >>> tuple(flatmap(['one two', 'three four', 'five'], str.split))
    ('one', 'two', 'three', 'four', 'five')
    '''
    return itertools_chain.from_iterable(builtin_map(func, iterable))


@export_from_module
def flatten(iterable):
    '''
    Assuming iterables consists from another iterables, yields items from "sub-iterables" one-by-one.

    :param iterable:
    :return:

    Example:
    >>> tuple(flatten(((1, 2), (3,))))
    (1, 2, 3)
    '''
    return itertools_chain.from_iterable(iterable)


@export_from_module
def intersperse(iterable, item):
    '''
    Returns iterator where `item` is placed between original items.

    :param iterable:
    :param item:
    :return:

    Example:
    >>> tuple(intersperse([1, 2, 3], 0))
    (1, 0, 2, 0, 3)
    '''
    # TODO: think about better solution
    it = iter(iterable)
    with contextlib_suppress(StopIteration):
        yield next(it)
        for elem in it:
            yield item
            yield elem


@export_from_module
def map(iterable, func):
    '''
    Returns iterator where `func` is applied to every item in `iterable`.

    :param iterable:
    :param func:
    :return:

    Example:
    >>> tuple(map('abc', str.upper))
    ('A', 'B', 'C')
    '''
    return builtin_map(func, iterable)


@export_from_module
@alias_for(drop)
def skip(iterable, number=1):
    pass


@export_from_module
@alias_for(drop_while)
def skip_while(iterable, predicate):
    pass


@export_from_module
@alias_for(itertools_islice)
def slice(iterable, *args):
    # TODO: make it work for all variants of `start`, `stop`, `step`, e.g. negative ones. See more-itertools.
    # Note: don't forget to remove @alias_for
    pass


@export_from_module
def starmap(iterable, func):
    '''
    Returns iterable where `func` is applied to every item in `iterable`, assuming that items are iterables with common length and `func`
    is a function of that length arguments.

    :param iterable:
    :param func:
    :return:

    Example:
    >>> tuple(starmap(('ab', 'ac', 'za'), lambda f, s: f > s))
    (False, False, True)
    '''
    return itertools_starmap(func, iterable)


@export_from_module
def step(iterable, k=1):
    '''
    Returns an iterator that yields every `k`th item from `iterable`.

    :param iterable:
    :param k:
    :return:

    Example:
    >>> tuple(step('abcdef', 3))
    ('a', 'd')
    '''
    return itertools_islice(iterable, None, None, k)


@export_from_module
def take(iterable, number=1):
    '''
    Returns an iterable that yields at *most* `number` first items from `iterable`.
    "Taken" items are not consumed from original iterable after function call.

    :param iterable:
    :param number:
    :return:

    Example:
    >>> tuple(take(range(100000000), 3))
    (0, 1, 2)
    '''
    return itertools_islice(iterable, number)


@export_from_module
def take_now(iterable, number=1):
    '''
    Returns a tuple with at *most* `number` first items from `iterable`.

    :param iterable:
    :param number:
    :return:

    Example:
    >>> take_now(range(1000000000), 3)
    (0, 1, 2)
    '''
    return tuple(itertools_islice(iterable, number))


@export_from_module
def take_last(iterable, number=1):
    '''
    Returns a tuple with at *most* `number` last items from `iterable`.

    :param iterator:
    :param number:
    :return:

    Example:
    >>> take_last('abcdef', 3)
    ('d', 'e', 'f')
    '''
    return tuple(collections_deque(iterable, maxlen=number))


@export_from_module
def take_while(iterable, predicate):
    '''
    Returns an iterable that yields original items while `predicate` is True. "Taken" items are not consumed from original iterable
    after function call.
    Note: first `predicate`-negative item will be consumed (see Example section).

    :param iterator:
    :param predicate:
    :return:

    Example:
    >>> it = iter('abcDef'); tuple(take_while(it, str.islower)), tuple(it)
    (('a', 'b', 'c'), ('e', 'f'))
    '''
    return itertools_takewhile(predicate, iterable)




import unittest

from .extra import TestExtraAlgorithms
from .folding import TestFoldingAlgorithms
from .merging import TestMergingAlgorithms
from .predicate_on_sequence import TestPredicateOnSequenceAlgorithms
from .searching import TestSearchingAlgorithms
from .sequence_to_groups import TestSequenceToGroupsAlgorithms
from .sequence_to_sequence import TestSequenceToSequenceAlgorithms
from .splitting import TestSplittingAlgorithms


def get_algorithms_test_suite():
    suite = unittest.TestSuite()
    suite.addTests(
        map(
            unittest.defaultTestLoader.loadTestsFromTestCase, (
                TestExtraAlgorithms,
                TestFoldingAlgorithms,
                TestMergingAlgorithms,
                TestPredicateOnSequenceAlgorithms,
                TestSearchingAlgorithms,
                TestSequenceToGroupsAlgorithms,
                TestSequenceToSequenceAlgorithms,
                TestSplittingAlgorithms
            )
        )
    )
    return suite

__all__ = ['get_algorithms_test_suite']



from abc import ABC, abstractmethod
# from iter2.iterators.iterator2 import Iterator2


class Option2(ABC):
    def __getattr__(self, item):
        return getattr(iter(self), item)

    @abstractmethod
    def __bool__(self):
        pass

    @abstractmethod
    def expect(self, exception_type, *ex_args, **ex_kwargs):
        pass

    @abstractmethod
    def is_some(self):
        pass

    @abstractmethod
    def is_none(self):
        pass

    @abstractmethod
    def unwrap(self):
        pass

    @abstractmethod
    def unwrap_or(self, default):
        pass

    @abstractmethod
    def unwrap_or_else(self, default_func):
        pass

    @abstractmethod
    def map(self, func):
        pass

    @abstractmethod
    def map_or(self, default, func):
        pass

    @abstractmethod
    def map_or_else(self, default_func, func):
        pass

    @abstractmethod
    def __and__(self, other):
        pass

    def and_(self, other):
        return self & other

    @abstractmethod
    def and_then(self, func):
        pass

    @abstractmethod
    def filter(self, predicate):
        pass

    @abstractmethod
    def __or__(self, other):
        pass

    def or_(self, other):
        return self | other

    @abstractmethod
    def or_else(self, func):
        pass

    @abstractmethod
    def __xor__(self, other):
        pass

    def xor(self, other):
        return self ^ other


class None2Type(Option2):
    def __repr__(self):
        return '.'.join((self.__module__ , 'None2'))

    def __iter__(self):
        return iter(())

    def __bool__(self):
        return False

    def is_some(self):
        return False

    def is_none(self):
        return True

    def expect(self, exception_type, *ex_args, **ex_kwargs):
        raise exception_type(*ex_args, **ex_kwargs)

    def unwrap(self):
        # TODO: think about better exception type
        raise RuntimeError('Error unwrapping None')

    def unwrap_or(self, default):
        return default

    def unwrap_or_else(self, default_func):
        return default_func()

    def map(self, func):
        return self

    def map_or(self, default, func):
        return default

    def map_or_else(self, default_func, func):
        return default_func()

    def filter(self, predicate):
        return self

    def __and__(self, other):
        return self

    def and_then(self, func):
        return self

    def __or__(self, other):
        return other

    def or_else(self, func):
        return func()

    def __xor__(self, other):
        if other.is_some():
            return other
        else:
            return self


None2 = None2Type()  # Only one instance


class Some2(Option2):
    __slots__ = ('_value',)

    def __init__(self, value):
        self._value = value

    def __repr__(self):
        return ''.join(('.'.join((self.__module__, 'Some2')), '(', repr(self._value), ')'))

    def __iter__(self):
        # return Iterator2((self._value,))
        yield self._value

    def __bool__(self):
        return True

    def is_some(self):
        return True

    def is_none(self):
        return False

    def expect(self, exception_type, *ex_args, **ex_kwargs):
        return self._value

    def unwrap(self):
        return self._value

    def unwrap_or(self, default):
        return self._value

    def unwrap_or_else(self, default_func):
        return self._value

    def map(self, func):
        return Some2(func(self._value))

    def map_or(self, default, func):
        return func(self._value)

    def map_or_else(self, default_func, func):
        return func(self._value)

    def filter(self, predicate):
        if predicate(self._value) is True:
            return Some2(self._value)
        else:
            return None2

    def __and__(self, other):
        return other

    def and_then(self, func):
        return func()

    def __or__(self, other):
        return self

    def or_else(self, func):
        return self

    def __xor__(self, other):
        if other.is_none():
            return self
        else:
            return None2


from .iterators.iterator2 import Iterator2
from . import option
from . import sequence_builders
from . import wrapped_algorithms

from iter2.utils import define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup exports


@export_from_module
def iter2(*args):
    return Iterator2(*args)


# Populate option
iter2.some = option.Some2
iter2.none = option.None2


# Populate sequence builders
iter2.sequence_builders = sequence_builders
iter2.count_from = sequence_builders.count_from
iter2.cycle = sequence_builders.cycle
iter2.empty = sequence_builders.empty
iter2.from_args = sequence_builders.from_args
iter2.of = iter2.from_args
iter2.iterate = sequence_builders.iterate
iter2.numeric_range = sequence_builders.numeric_range
iter2.once = sequence_builders.once
iter2.range = sequence_builders.range
iter2.repeat = sequence_builders.repeat
iter2.repeat_call = sequence_builders.repeat_call
iter2.tabulate = sequence_builders.tabulate


# Populate wrapped algorithms
iter2.cartesian_product = wrapped_algorithms.castesian_product
iter2.sort_together = wrapped_algorithms.sort_together
iter2.unique_to_each = wrapped_algorithms.unique_to_each
iter2.chain = wrapped_algorithms.chain
iter2.chain_from_iterable = wrapped_algorithms.chain_from_iterable
iter2.interleave = wrapped_algorithms.interleave
iter2.interleave_shortest = wrapped_algorithms.interleave_shortest
iter2.roundrobin = wrapped_algorithms.roundrobin
iter2.zip = wrapped_algorithms.zip
iter2.zip_longest = wrapped_algorithms.zip_longest
iter2.zip_offset = wrapped_algorithms.zip_offset




import itertools

from iter2.utils import alias_for, define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup export


# Aliases
builtin_zip = zip

itertools_chain = itertools.chain
itertools_chain_from_iterable = itertools.chain.from_iterable
itertools_cycle = itertools.cycle
itertools_islice = itertools.islice
itertools_repeat = itertools.repeat
itertools_zip_longest = itertools.zip_longest


@export_from_module
@alias_for(itertools_chain)
def chain(*iterables):
    '''
    Returns a new iterable yielding items from `iterables`, iterable by iterable.

    :param iterables:
    :return:  chained iterables as new iterable

    Example:
    >>> list(chain([1, 2], 'ab'))
    [1, 2, 'a', 'b']

    See also `chain_from_iterable`
    '''
    pass


@export_from_module
@alias_for(itertools_chain_from_iterable)
def chain_from_iterable(iterables_in_iterable):
    '''
    Returns a new iterable yielding items from iterables, iterable by iterable.
    This is the same as `chain` b   ut consumes iterables from another iterable.

    :param iterables_in_iterable:
    :return:  chained iterables as new iterable

    Example:
    >>> list(chain_from_iterable(['ab', 'cd', 'ef']))
    ['a', 'b', 'c', 'd', 'e', 'f']
    '''
    pass


@export_from_module
def interleave(*iterables):
    '''
    Returns a new iterable yielding from each iterable in turn, until the longest is exhausted.

    :param iterables:
    :return:

    Example:
    >>> interleave([1, 2], 'abc')
    [1, 'a', 2, 'b', 'c']
    '''
    # TODO: think about better solution
    MISSING_MARK = object()  # truly unique value
    return (
        item
        for item in itertools_chain_from_iterable(itertools_zip_longest(*iterables, fillvalue=MISSING_MARK))
        if item is not MISSING_MARK
    )


@export_from_module
def interleave_shortest(*iterables):
    '''
    Returns a new iterable yielding from each iterable in turn, until the shortest is exhausted.

    :param iterables:
    :return:  iterable

    Example:
    >>> interleave_shortest([1, 2], 'abc')
    [1, 'a', 2, 'b']
    '''
    # TODO: think about better solution
    return itertools_chain_from_iterable(builtin_zip(*iterables))


@export_from_module
def prepend(iterator, another_iterator):
    '''
    Returns a new iterable, alias for `chain(another_iterator, iterator)`.
    First, yields items from `another_iterator`, then from `iterator`.

    :param iterator:
    :param another_iterator:
    :return:  iterable

    Example:
    >>> tuple(prepend([4, 5, 6], [1, 2, 3]))
    (1, 2, 3, 4, 5, 6)
    '''
    return itertools_chain(another_iterator, iterator)


@export_from_module
@alias_for(interleave)
def roundrobin(*iterables):
    pass


@export_from_module
@alias_for(builtin_zip)
def zip(*iterables):
    pass


@export_from_module
@alias_for(itertools_zip_longest)
def zip_longest(*iterables, fillvalue=None):
    pass


@export_from_module
def zip_offset(*iterables, offsets, fillvalue=None, longest=False):
    '''
    `zip` the input `iterables` together, but offset the i-th iterable by the i-th item in `offsets`.

    :param iterables:
    :param offsets:
    :param fillvalue:  fillvalue for longest=True case
    :param longest:  if `True` iterates until longest iterable exhausted
    :return:  tuples with zipped items

    Example:
    >>> tuple(zip_offset('123', 'abcde', offsets=(0, 1)))
    (('1', 'b'), ('2', 'c'), ('3', 'd'))
    >>> tuple(zip_offset('123', 'abcde', offsets=(0, 1), fillvalue='#', longest=True))
    (('1', 'b'), ('2', 'c'), ('3', 'd'), ('#', 'e'))
    '''
    # based on more-itertools::zip_offset
    if len(iterables) != len(offsets):
        raise ValueError("Number of iterables and offsets didn't match")

    staggered = []
    add_to_staggered = staggered.append
    for it, n in builtin_zip(iterables, offsets):
        if n < 0:
            add_to_staggered(
                itertools_chain(
                    itertools_repeat(fillvalue, -n),
                    it
                )
            )
        elif n > 0:
            add_to_staggered(itertools_islice(it, n, None))
        else:
            add_to_staggered(it)

    if longest:
        return itertools_zip_longest(*staggered, fillvalue=fillvalue)
    else:
        return builtin_zip(*staggered)




import itertools
import collections

itertools_chain = itertools.chain
itertools_islice = itertools.islice
collections_deque = collections.deque


# TODO: refactor this shit
class Result(object):
    __slots__ = ('value',)

    def __init__(self, value=None):  # assumed not used explicitly
        self.value = value

    def __iter__(self):
        yield self.is_ok  # for child classes only
        yield self.value


class Ok(Result):
    is_ok = True
    is_error = False


class Error(Result):
    is_ok = False
    is_error = True

#


class SpyIterator(object):
    __slots__ = ('_iterator', '_spied')

    def __init__(self, iterable):
        self._iterator = iter(iterable)
        self._spied = collections_deque()

    def __iter__(self):
        spied = self._spied
        for _ in range(len(spied)):
            yield spied.popleft()
        yield from self._iterator

    def __next__(self):
        spied = self._spied
        if len(spied) > 0:
            return spied.popleft()
        else:
            return next(self._iterator)

    def raw(self):
        raw_it = itertools_chain(self._spied, self._iterator)
        self.__init__(())
        return raw_it

    def spy(self, n=1, *, allow_partial=False):
        spied = self._spied
        spied_count = len(spied)
        if n > spied_count:
            spied.extend(itertools_islice(self._iterator, n - spied_count))
            if n > len(spied):
                if not allow_partial or n == 1:
                    return Error()
                else:
                    return Ok(tuple(itertools_islice(spied, n)))
        if n == 1:
            return Ok(spied[0])
        else:
            return Ok(tuple(itertools_islice(spied, n)))



from iter2.utils import define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup export


builtin_enumerate = enumerate
builtin_filter = filter


@export_from_module
def find(iterable, predicate):
    '''
    Yields first occurrence of item for which `predicate` is True

    :param iterable:
    :param predicate:
    :return:

    Example:
    >>> find([1, 2, 3], lambda x: x % 2 == 0)
    2
    '''
    return next(builtin_filter(predicate, iterable))


@export_from_module
def locate(iterable, predicate, *, count_from=0, with_items=False):
    '''
    Searches items for which `predicate` is True. Yields indexes started from `count_from`.
    If `with_items` is True then yields tuples (index, item)

    :param iterable:
    :param predicate:
    :param count_from:
    :param with_items:
    :return:  indexes  or  (index, item),...

    Example:
    >>> tuple(locate([1, 2, 3, 4, 5], lambda x: x % 2 == 0, count_from=1))
    (2, 4)
    >>> tuple(locate(['aBcDe'], str.isupper, with_items=True))
    ((1, 'B'), (3, 'D'))
    '''

    if with_items:
        return (
            (idx, item)
            for idx, item in builtin_enumerate(iterable, count_from)
            if predicate(item)
        )
    else:
        return (
            idx
            for idx, item in builtin_enumerate(iterable, count_from)
            if predicate(item)
        )


@export_from_module
def position(iterable, predicate, *, count_from=0):
    '''
    Returns index (starting from `count_from`) of first occurrence of item for which `predicate` is True.
    If `predicate` is False for all items returns -1.

    :param iterable:
    :param predicate:
    :param count_from:
    :return:  index   or  -1

    Example:
    >>> position('aBcD', str.isupper, count_from=1)
    2
    >>> position('abcd', str.isupper, count_from=1)
    -1
    '''
    return next(locate(iterable, predicate, count_from=count_from), -1)




import itertools
import operator

from iter2.iterators.spy_iterator import SpyIterator
from iter2.utils import define_module_exporter, Flag


export_from_module, __all__ = define_module_exporter()  # setup export


# Aliases
builtin_enumerate = enumerate
builtin_filter = filter
builtin_map = map

itertools_filterfalse = itertools.filterfalse
itertools_islice = itertools.islice
itertools_tee = itertools.tee

operator_itemgetter = operator.itemgetter


@export_from_module
def distribute(iterable, n):
    '''
    Distributes items from `iterable` among `n` smaller iterables.

    :param iterable:
    :param n:
    :return:

    Example:
    >>> tuple(map(tuple, distribute([1, 2, 3, 4, 5], 3)))
    ((1, 4), (2, 5), (3,))
    '''
    copies = itertools_tee(iterable, n)
    return tuple(
        itertools_islice(copy, idx, None, n)
        for idx, copy in builtin_enumerate(copies)
    )


@export_from_module
def partition(iterable, predicate):
    '''
    Splits items from `iterable` into two iterators w/ `predicate`-positive and `predicate`-negative elements.

    :param iterable:
    :param predicate:
    :return:

    Example:
    >>> tuple(map(tuple, partition('AbC', str.isupper)))
    (('A', 'C'), ('b',))
    '''
    copy1, copy2 = itertools_tee(iterable, 2)
    positive = builtin_filter(predicate, copy1)
    negative = itertools_filterfalse(predicate, copy2)
    return positive, negative


@export_from_module
def split_after(iterable, predicate):
    '''
    Splits `iterable` after every `predicate`-positive item into sub-iterators.

    :param iterable:
    :param predicate:
    :return:

    Example:
    >>> tuple(map(''.join, split_after("Ok!Let's go!", lambda c: c == '!')))
    ('Ok!', "Let's go!")
    '''
    spy_it = SpyIterator(iterable)

    def split_chunk_gen(it, pred):
        for item in it:
            if pred(item):
                yield item
                return
            else:
                yield item

    while spy_it.spy().is_ok:
        yield split_chunk_gen(spy_it, predicate)


@export_from_module
def split_at(iterable, predicate):
    '''
    Splits `iterable` on every `predicate`-positive item into sub-iterators. Delimiting items are not included.

    :param iterable:
    :param predicate:
    :return:

    Example:
    >>> tuple(map(tuple, split_at([1, 2, 0, 4, 5, 0], bool)))
    ((1, 2), (4, 5), ())
    '''
    spy_it = SpyIterator(iterable)
    previously_splitted = Flag()

    def split_chunk_gen(it, pred):
        for item in it:
            if pred(item):
                previously_splitted.set()
                return
            else:
                yield item

    while spy_it.spy().is_ok or previously_splitted:
        previously_splitted.clear()
        yield split_chunk_gen(spy_it, predicate)


@export_from_module
def split_before(iterable, predicate):
    '''
    Splits `iterable` before every `predicate`-positive item into sub-iterators.

    :param iterable:
    :param predicate:
    :return:

    Example:
    >>> tuple(map(''.join, split_before("Ok!Let's go!", str.isupper)))
    ('Ok!', "Let's go!")
    '''
    spy_it = SpyIterator(iterable)

    def split_chunk_gen(spy_it, pred):
        ok, item = spy_it.spy()
        if pred(item):
            next(spy_it) # skip first if `pred`-positive
            yield item
            ok, item = spy_it.spy()

        while ok and not pred(item):
            next(spy_it)  # consume one
            yield item
            ok, item = spy_it.spy()

    while spy_it.spy().is_ok:
        yield split_chunk_gen(spy_it, predicate)


@export_from_module
def unzip(iterable, arity=2):
    '''
    Produces `arity`-count of iterators from `iterable`. i-th iterator yields every i-th element from `arity`-length item.

    :param iterable:
    :param arity:
    :return:

    Example:
    >>> tuple(map(tuple, unzip([(1, 2, 3), (4, 5, 6)], arity=3)))
    ((1, 4), (2, 5), (3, 6))
    '''
    if not isinstance(arity, int) or arity < 2:
        raise TypeError('`arity` must be integer greater or equal to 2')
    if arity == 2:
        copy1, copy2 = itertools_tee(iterable, 2)
        first_flow = (first for first, _ in copy1)
        second_flow = (second for _, second in copy2)
        return first_flow, second_flow
    else:
        return tuple(
            (builtin_map(operator_itemgetter(idx), copy)
             for idx, copy in builtin_enumerate(itertools_tee(iterable, arity)))
        )



import functools
import operator

from iter2.utils import alias_for, define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup export


# Aliases
builtin_sum = sum
builtin_max = max
builtin_min = min

functools_reduce = functools.reduce
operator_mul = operator.mul


@export_from_module
def count(iterable):
    '''
    Counts number of items in iterable.

    :param iterable:
    :return:  number of items in `iterable`

    Example:
    >>> count([1, 2, 3])
    3
    '''
    return builtin_sum(1 for _ in iterable)


@export_from_module
def fold(iterable, func, *, initial=None):
    '''
    Apply a function `func` of two arguments cumulatively to the items of a `iterable`,
    from left to right, so as to reduce the sequence to a single value. If `initial` is not None
    then it is used as the first state value.

    :param iterable:
    :param func:  (state_value, next_item) -> new_state_value
    :param initial:  optional first state value
    :return:  cumulative value

    Example:
    >>> fold([1, 2, 3, 4, 5], lambda state, item: state * item)
    120
    '''
    if initial is None:
        return functools_reduce(func, iterable)
    else:
        return functools_reduce(func, iterable, initial)


@export_from_module
def join(iterable, sep):
    '''
    Concatenates strings from `iterator` with `sep` between adjacent items.

    :param iterable:
    :param sep:   separator
    :return:  concatenated string

    Example:
    >>> join(['555', '404', '42', '73'], '-')
    '555-404-42-73'
    '''
    # TODO: use only when type hints are available
    return sep.join(iterable)


@export_from_module
def max(iterable, *, default=None, key=None):
    '''
    Returns biggest item from iterable. If iterable is empty returns None.
    If `key` function is set then `key(item)` is used as value for comparison.

    :param iterable:
    :param default:  default value to return if iterable is empty
    :param key:  function to calculate comparison key
    :return:  biggest item

    Example:
    >>> max([1, 2, 3])
    3
    >>> max([1, 2, 3], key=lambda x: x % 3)
    2
    >>> max([], default='nothing')
    'nothing'
    '''
    if key is None:
        return builtin_max(iterable, default=default)
    else:
        return builtin_max(iterable, default=default, key=key)


@export_from_module
def min(iterator, *, default=None, key=None):
    '''
    Returns smallest item from iterable. If iterable is empty returns None.
    If `key` function is set then `key(item)` is used as value for comparison.

    :param iterator:
    :param default:  default value to return if iterable is empty
    :param key:  function to calculate comparison key
    :return:  smallest value

    Example:
    >>> min([1, 2, 3])
    1
    >>> min([1, 2, 3], key=lambda x: x % 3)
    3
    >>> min([], default='nothing')
    'nothing'
    '''
    if key is None:
        return builtin_min(iterator, default=default)
    else:
        return builtin_min(iterator, default=default, key=key)


@export_from_module
def minmax(iterable, *, default=None, key=None):
    '''
    Returns smallest and biggest item from iterable. If iterable is empty returns `None`.
    If `key` function is set then `key(item)` is used as value for comparison.

    :param iterable:
    :param default:  default value to return if iterable is empty
    :param key:  function to calculate comparison key
    :return:  (smallest item, biggest item)

    Example:
    >>> minmax([1, 2, 3])
    (1, 3)
    >>> minmax([1, 2, 3], key=lambda x: x % 3)
    (3, 2)
    >>> minmax([], default='nothing')
    'nothing'
    '''
    # TODO: search for more optimized algo
    iterator = iter(iterable)
    if key is None:  # w/o key
        try:
            first = next(iterator)
            min_element = first
            max_element = first
        except StopIteration:
            return default

        for item in iterator:
            if item < min_element:
                min_element = item
            elif item > max_element:
                max_element = item

        return min_element, max_element

    else:
        try:
            first = next(iterator)
            first_key = key(first)
            min_element, min_key = first, first_key
            max_element, max_key = first, first_key
        except StopIteration:
            return default

        for item in iterator:
            item_key = key(item)
            if item_key < min_key:
                min_element, min_key = item, item_key
            elif item_key > max_key:
                max_element, max_key = item, item_key

        return min_element, max_element


@export_from_module
def product(iterable):
    '''
    Returns the product of items from `iterable`.

    :param iterable:
    :return:  product of items

    Example:
    >>> product([1, 2, 3, 4, 5])
    120
    '''
    return functools_reduce(operator_mul, iterable)


@export_from_module
@alias_for(fold)
def reduce(iterable, func, *, initial=None):
    pass


@export_from_module
def sum(iterable):
    ''''
    Returns the sum of items from `iterable`.

    :param iterable:
    :return:  sum of items

    Example:
    >>> sum([1, 2, 3, 4, 5])
    15
    '''
    return builtin_sum(iterable)



import unittest

import iter2.algorithms.splitting as algo


class TestSplittingAlgorithms(unittest.TestCase):
    def assertEqualAfterMaterialize(self, first, second):
        self.assertEqual(
            tuple(map(tuple, first)),
            tuple(second)
        )

    def test_distribute(self):
        with self.subTest('simple'):
            self.assertEqualAfterMaterialize(
                algo.distribute([1, 2, 3, 4, 5], 3),
                ((1, 4), (2, 5), (3,))
            )

    def test_partition(self):
        with self.subTest('simple'):
            self.assertEqualAfterMaterialize(
                algo.partition('AbCd', str.isupper),
                (('A', 'C'), ('b', 'd'))
            )

    def test_split_after(self):
        with self.subTest('simple'):
            self.assertEqualAfterMaterialize(
                algo.split_after('192.168.0.1', lambda c: c == '.'),
                map(tuple, ('192.', '168.', '0.', '1'))
            )

    def test_split_at(self):
        with self.subTest('simple'):
            self.assertEqualAfterMaterialize(
                algo.split_at('192.168.0.1', lambda c: c == '.'),
                map(tuple, ('192', '168', '0', '1'))
            )

    def test_split_before(self):
        with self.subTest('simple'):
            self.assertEqualAfterMaterialize(
                algo.split_before('192.168.0.1', lambda c: c == '.'),
                map(tuple, ('192', '.168', '.0', '.1'))
            )

    def test_unzip(self):
        with self.subTest('simple'):
            self.assertEqualAfterMaterialize(
                algo.unzip([(1, 2), (3, 4)]),
                ((1, 3), (2, 4))
            )

        with self.subTest('custom arity'):
            self.assertEqualAfterMaterialize(
                algo.unzip([(1, 2, 3), (4, 5, 6)], arity=3),
                ((1, 4), (2, 5), (3, 6))
            )



from functools import wraps

from .iterators.iterator2 import Iterator2

from .algorithms import extra
from .algorithms import merging


def wraps_algo(func):
    def wraps_algo_func(dummy):
        func.__doc__ = func.__doc__ + \
            '\n---\nNote: This version wraps result into `Iterator2`'
        @wraps(func)
        def wrapper(*args, **kwargs):
            return Iterator2(func(*args, **kwargs))
        return wrapper
    return wraps_algo_func


@wraps_algo(extra.cartesian_product)
def castesian_product(*iterables, repeat=1):
    pass


@wraps_algo(extra.sort_together)
def sort_together(*iterables, key_list=(0,), key_func=None, reverse=False):
    pass


@wraps_algo(extra.unique_to_each)
def unique_to_each(*iterables):
    pass


@wraps_algo(merging.chain)
def chain(*iterables):
    pass


@wraps_algo(merging.chain_from_iterable)
def chain_from_iterable(iterable):
    pass


@wraps_algo(merging.interleave)
def interleave(*iterables):
    pass


@wraps_algo(merging.interleave_shortest)
def interleave_shortest(*iterables):
    pass


@wraps_algo(merging.roundrobin)
def roundrobin(*iterables):
    pass


@wraps_algo(merging.zip)
def zip(*iterables):
    pass


@wraps_algo(merging.zip_longest)
def zip_longest(*iterables, fillvalue=None):
    pass


@wraps_algo(merging.zip_offset)
def zip_offset(*iterables, offsets, fillvalue=None, longest=None):
    pass




import itertools
import operator
import collections

from . import merging
from iter2.utils import define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup export


# Aliases
builtin_enumerate = enumerate
builtin_filter = filter
builtin_map = map
builtin_zip = zip

itertools_groupby = itertools.groupby
itertools_islice = itertools.islice
itertools_permutations = itertools.permutations
itertools_tee = itertools.tee
itertools_zip_longest = itertools.zip_longest

collections_deque = collections.deque
operator_itemgetter = operator.itemgetter

merging_zip_offset = merging.zip_offset


@export_from_module
def chunks(iterable, size, *, allow_partial=False):
    '''
    Groups items in chunks of size `size`. If `allow_partial` is True last chunk may be incomplete.

    :param iterable:
    :param size:  length of chunk
    :param allow_partial:
    :return:  tuples of size `size`

    Example:
    >>> tuple(chunks([1, 2, 3, 4], 2))
    ((1, 2), (3, 4))
    >>> tuple(chunks([1, 2, 3], 2, allow_partial=True))
    ((1, 2), (3,))
    '''
    if allow_partial is False:
        it_copies = (iter(iterable),) * size
        yield from builtin_zip(*it_copies)
    else:
        it = iter(iterable)
        next_diff_piece = tuple(itertools_islice(it, size))
        while len(next_diff_piece) == size:
            yield next_diff_piece
            next_diff_piece = tuple(itertools_islice(it, size))
        if len(next_diff_piece) > 0:
            yield next_diff_piece


@export_from_module
def chunks_with_padding(iterable, size, *, fillvalue=None):
    '''
    Groups items in chunks of size `size`. Fills incomplete chunk with `fillvalue`.

    :param iterable:
    :param size:
    :param fillvalue:
    :return:  tuples of size `size`

    Example:
    >>> tuple(chunks_with_padding([1, 2, 3, 4], 2))
    ((1, 2), (3, 4))
    >>> tuple(chunks_with_padding([1, 2, 3], 2, fillvalue=100500))
    ((1, 2), (3, 100500))
    '''
    it_copies = (iter(iterable),) * size
    return itertools_zip_longest(*it_copies, fillvalue=fillvalue)


@export_from_module
def consecutive_groups(iterable, *, ordering=None):
    '''
    Groups items which are consecutive due to `ordering`. If `ordering` is None then it is assumed that items are integers.

    :param iterable:
    :param ordering:
    :return: tuples of consecutive groups

    Example:
    >>> tuple(consecutive_groups([1, 2, 10, 11, 100, 101]))
    ((1, 2), (10, 11), (100, 101))
    >>> tuple(consecutive_groups('abcBCDcde', ordering=ord))
    (('a', 'b', 'c'), ('B', 'C', 'D'), ('c', 'd', 'e'))
    '''
    if ordering is None:
        key = lambda idx, item: idx - item
    else:
        key = lambda idx, item: idx - ordering(item)
    return (
        builtin_map(operator_itemgetter(1), sub_iter)
        for _, sub_iter in itertools_groupby(builtin_enumerate(iterable), key=(lambda item: key(*item)))
    )


@export_from_module
def group_by(iterable, *, key=None):
    '''
    Groups items in consecutive keys and groups from the `iterable`.
    If `key` function is not specified or is None, the element itself is a key for grouping.

    :param iterable:
    :param key:
    :return:

    Example:
    >>> for k, group in group_by([1, 2, 2, 3, 3, 3], key=lambda x: x % 2):
    ...     print((k, tuple(group)))
    (1, (1,))
    (0, (2, 2))
    (1, (3, 3, 3))
    '''
    return itertools_groupby(iterable, key=key)



@export_from_module
def pairwise(iterable):
    '''
    Groups items in overlapping pairs from the original.

    :param iterable:
    :return:

    Example:
    >>> tuple(pairwise([1, 2, 3]))
    ((1, 2), (2, 3))
    '''
    copy1, copy2 = itertools_tee(iterable, 2)
    next(copy2, None)
    return builtin_zip(copy1, copy2)


@export_from_module
def permutations(iterable, length=None):
    '''
    Return successive `length`-permutations of items in the `iterable`.

    :param iterable:
    :param length:
    :return:

    Example:
    >>> tuple(permutations([0, 1, 2], length=2))
    ((0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1))
    '''
    return itertools_permutations(iterable, r=length)


@export_from_module
def process_in_groups(iterable, *, key=None, transformation=None, aggregator=None):
    '''
    Groups items in consecutive keys and groups from the `iterable`.
    If `key` function is not specified or is None, the element itself is a key for grouping.
    If `transformation` is specified then it will be applied to each item.
    If `aggregator` is specified then it will be applied to each group.

    :param iterable:
    :param key:
    :param transformation:
    :param aggregator:
    :return:

    Example:
    >>> for k, group in process_in_groups([1, 2, 2, 3, 3, 3], key=lambda x: x % 2):
    ...     print((k, tuple(group)))
    (1, (1,))
    (0, (2, 2))
    (1, (3, 3, 3))
    >>> for k, group in process_in_groups([1, 2, 2, 3, 3, 3], transformation=lambda x: x % 2):
    ...     print((k, tuple(group)))
    (1, (1,))
    (2, (0, 0))
    (3, (1, 1, 1))
    >>> for k, val in process_in_groups([1, 2, 2, 3, 3, 3], aggregator=sum):
    ...     print((k, val))
    (1, 1)
    (2, 4)
    (3, 9)
    '''
    result = itertools_groupby(iterable, key=key)
    if transformation is not None:
        result = (
            (val, builtin_map(transformation, sub_iter))
            for val, sub_iter in result
        )
    if aggregator is not None:
        result = (
            (val, aggregator(sub_iter))
            for val, sub_iter in result
        )
    return result


@export_from_module
def sliding_window(iterable, *, size=1, step=1, allow_partial=False):
    '''
    Sliding window (#docs_in_pictures, size=5, step=4)
    iterable:   a b c d e f g h i j k
               [         ] <- "window" (step 0)
               <- size ->
    res[0]:    (a,b,c,d,e)
    iterable:   a b c d e f g h i j k
                       [         ] <- "window" (step 1)
               <-step->
    res[1]:    (e, f, g, h, i)
    iterable:   a b c d e f g h i j k
                               [         ] <- "window" (step 2)
    res[2]:    (i, j, k)   # if `allow_partial` is True

    :param iterable:
    :param step:
    :param size:
    :param allow_partial:
    :return:
    '''
    if step == size:
        yield from chunks(iterable, size, allow_partial=allow_partial)
        return  # done

    # TODO: think about optimizing iteration in chunks

    it = iter(iterable)
    first_piece = tuple(itertools_islice(it, size))
    if len(first_piece) != size and not allow_partial:
        return  # not enough elements
    yield first_piece

    if step < size:
        dq = collections_deque(first_piece, maxlen=size)
        next_diff_piece = tuple(itertools_islice(it, step))
        while len(next_diff_piece) == step:
            dq.extend(next_diff_piece)
            yield tuple(dq)
            next_diff_piece = tuple(itertools_islice(it, step))
        if allow_partial and len(next_diff_piece) > 0:
            yield next_diff_piece
    else:  # step > size
        next_diff_piece = tuple(itertools_islice(it, step))
        while len(next_diff_piece) == step:
            yield next_diff_piece[-size:]
            next_diff_piece = tuple(itertools_islice(it, step))
        if allow_partial and len(next_diff_piece) > 0:
            idx_of_start = len(next_diff_piece) - (step - size)
            yield next_diff_piece[idx_of_start:]


@export_from_module
def stagger(iterable, *, offsets=(-1, 0, 1), fillvalue=None, longest=False):
    '''
    Yields tuples whose elements are offset from iterable.
    The amount by which the i-th item in each tuple is offset is given by the i-th item in `offsets`.
    >>> tuple(stagger([0, 1, 2, 3]))
    ((None, 0, 1), (0, 1, 2), (1, 2, 3))
    >>> tuple(stagger(range(8), offsets=(0, 2, 4)))
    ((0, 2, 4), (1, 3, 5), (2, 4, 6), (3, 5, 7))

    By default, the sequence will end when the final element of a tuple is the last item in the `iterable`.
    To continue until the first element of a tuple is the last item in the iterable, set `longest` to True:
    >>> tuple(stagger([0, 1, 2, 3], longest=True))
    ((None, 0, 1), (0, 1, 2), (1, 2, 3), (2, 3, None), (3, None, None))

    By default, None will be used to replace offsets beyond the end of the sequence.
    Specify `fillvalue` to use some other value.

    :param iterable:
    :param offsets:
    :param fillvalue:
    :param longest:
    :return:
    '''
    copies = itertools_tee(iterable, len(offsets))
    return merging_zip_offset(*copies, offsets=offsets, fillvalue=fillvalue, longest=longest)




from iter2.utils import define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup export


# Aliases
builtin_all = all
builtin_any = any
builtin_map = map


@export_from_module
def all(iterable, predicate):
    '''
    Checks if `predicate` is True on every item in `iterator`

    :param iterable:
    :param predicate:
    :return:  True or False

    Example:
    >>> all([1, 2, 3], lambda x: x < 10)
    True
    >>> all([1, 2, 3], lambda x: x % 2 == 0)
    False
    '''
    return builtin_all(builtin_map(predicate, iterable))


@export_from_module
def any(iterable, predicate):
    '''
    Checks if `predicate` is True on at least one item in `iterator`

    :param iterable:
    :param predicate:
    :return:  True or False

    Example:
    >>> any([1, 2, 3], lambda x: x > 10)
    False
    >>> any([1, 2, 3], lambda x: x % 2 == 0)
    True
    '''
    return builtin_any(builtin_map(predicate, iterable))


@export_from_module
def none(iterable, predicate):
    '''
    Checks if `predicate` is False on every item in `iterator`

    :param iterable:
    :param predicate:
    :return:  True or False

    Example:
    >>> none([1, 2, 3], lambda x: x < 10)
    True
    >>> none([1, 2, 3], lambda x: x % 2 == 0)
    False
    '''
    return not builtin_any(builtin_map(predicate, iterable))




import itertools
from functools import wraps

from iter2 import algorithms
from iter2.option import Some2, None2
from iter2.iterators.spy_iterator import SpyIterator

from iter2.utils import define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup export


# Shortcuts
algorithms_take = algorithms.sequence_to_sequence.take
algorithms_take_now = algorithms.sequence_to_sequence.take_now
algorithms_take_last = algorithms.sequence_to_sequence.take_last
algorithms_take_while = algorithms.sequence_to_sequence.take_while
algorithms_group_by = algorithms.sequence_to_groups.group_by
algorithms_process_in_groups = algorithms.sequence_to_groups.process_in_groups
algorithms_position = algorithms.searching.position

itertools_chain = itertools.chain
itertools_islice = itertools.islice
itertools_tee = itertools.tee


class InvalidIteratorClass(object):
    def __iter__(self):
        raise ValueError('Iterating invalidated iterator (probably original was moved).')

    def __next__(self):
        raise ValueError('Iterating invalidated iterator (probably original was moved).')


InvalidIterator = InvalidIteratorClass()


def merge_docs(method, algorithm_function):
    header = method.__doc__ or 'Derived from function `{fname}`'.format(fname=algorithm_function.__qualname__)
    return '\n---\n'.join([header, algorithm_function.__doc__])


def derived_from(algorithm_function, *, wrap_into_iterator=True):
    def derive_from_algo_func(method):
        method.__doc__ = merge_docs(method, algorithm_function)
        if wrap_into_iterator:
            @wraps(method)
            def wrapper(self, *args, **kwargs):
                return Iterator2(algorithm_function(self.raw(), *args, **kwargs))
        else:
            @wraps(method)
            def wrapper(self, *args, **kwargs):
                return algorithm_function(self.raw(), *args, **kwargs)
        return wrapper
    return derive_from_algo_func


def based_on(algorithm_function):
    def base_on_algo_func(method):
        method.__doc__ = merge_docs(method, algorithm_function)
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            self._iterator = algorithm_function(self._iterator, *args, **kwargs)  # patches inner iterator with algo
            return method(self, *args, **kwargs)
        return wrapper
    return base_on_algo_func


def with_rich_subiterators(*, in_tuple=False):
    def _with_rich_subiterators(method):
        if in_tuple:
            @wraps(method)
            def wrapper(self, *args, **kwargs):
                return tuple(map(Iterator2, self.raw()))
        else:
            @wraps(method)
            def wrapper(self, *args, **kwargs):
                return Iterator2(map(Iterator2, self.raw()))
        return wrapper
    return _with_rich_subiterators


@export_from_module
class Iterator2:
    __slots__ = ('_iterator',)

    def __init__(self, *args):
        self._iterator = iter(*args)

    def __repr__(self):
        return '{class_name}<{it_id}>'.format(
            class_name=self.__class__.__name__,
            it_id=hex(id(self))
        )

    # ====================
    # `iterator` behaviour
    # ====================
    def __iter__(self):
        return self  # Do not expose internals.

    def __next__(self):
        return next(self._iterator)  # Proxing next-protocol.

    def next(self):
        try:
            return Some2(next(self._iterator))
        except StopIteration:
            return None2

    def raw(self):
        '''
        Returns inner raw iterator and makes self invalid.

        :return:
        '''
        self._iterator, it = InvalidIterator, self._iterator  # move semantics
        return it

    def ref(self):
        '''
        Creates a copy-by-reference.
        Warning: Assumed to be used in special cases.

        :return:

        Example:
        >>> it = Iterator2(range(20))
        >>> it.ref().take_now(5)  # new Iterator2 that shares raw iterator w/ `it` and then consumes 5 items
        (0, 1, 2, 3, 4)
        >>> it.ref().take_now(5)  # new Iterator2 that also shares raw iterator w/ `it` and then consumes next 5 items
        (5, 6, 7, 8, 9)
        >>> it.take_now(5) #  `it` consumes 5 items and become invalid, {15, 16, 17, 18, 19} are "lost" here
        (10, 11, 12, 13, 14)
        >>> it.raw() is InvalidIterator
        True
        '''
        return Iterator2(self._iterator)

    # =================================
    # Processing "sequence -> sequence"
    # =================================
    @derived_from(algorithms.sequence_to_sequence.add_side_effect)
    def add_side_effect(self, func, chunk_size=None, before=None, after=None):
        pass

    @derived_from(algorithms.sequence_to_sequence.accumulate)
    def accumulate(self, func=None, *, initial=None):
        pass

    @derived_from(algorithms.sequence_to_sequence.consume)
    def consume(self, number=None):
        pass

    @derived_from(algorithms.sequence_to_sequence.cycle)
    def cycle(self, *, number=None):
        pass

    @derived_from(algorithms.sequence_to_sequence.dedup)
    def dedup(self):
        pass

    @derived_from(algorithms.sequence_to_sequence.difference)
    def difference(self, func=None, *, initial=None):
        pass

    @derived_from(algorithms.sequence_to_sequence.drop)
    def drop(self, number=1):
        pass

    @derived_from(algorithms.sequence_to_sequence.drop_while)
    def drop_while(self, predicate):
        pass

    @derived_from(algorithms.sequence_to_sequence.enumerate)
    def enumerate(self, *, count_from=0):
        pass

    @derived_from(algorithms.sequence_to_sequence.filter)
    def filter(self, predicate, *, inverse=False):
        pass

    @derived_from(algorithms.sequence_to_sequence.filter_none)
    def filter_builtin_none(self):
        pass

    def filter_none(self):
        '''
        Works with sequences of `Option`s. Unwraps `Some`s and filter `None`s.

        :return: new iterator
        Example:
        >>> tuple(Iterator2([Some2(1), None2, Some2(3)]).filter_none())
        (1, 3)
        '''
        self._iterator = (
            val.unwrap()
            for val in self._iterator
            if val.is_some()
        )
        return self

    @derived_from(algorithms.sequence_to_sequence.flatmap)
    def flatmap(self, func):
        pass

    @derived_from(algorithms.sequence_to_sequence.flatten)
    def flatten(self):
        pass

    def for_each(self, func):
        '''
        Calls`func` on every item. Returns nothing. Is equivalent to:
            for item in it:
                func(it)

        :param func:
        :return:  nothing

        Example:
        >>> Iterator2([1, 2]).for_each(print)
        1
        2
        '''
        for item in self._iterator:
            func(item)

    @derived_from(algorithms.sequence_to_sequence.intersperse)
    def intersperse(self, item):
        pass

    @derived_from(algorithms.sequence_to_sequence.map)
    def map(self, func):
        pass

    @derived_from(algorithms.sequence_to_sequence.skip)
    def skip(self, number=1):
        pass

    @derived_from(algorithms.sequence_to_sequence.skip_while)
    def skip_while(self, predicate):
        pass

    @derived_from(algorithms.sequence_to_sequence.slice)
    def slice(self, *args):
        pass

    @derived_from(algorithms.sequence_to_sequence.starmap)
    def starmap(self, func):
        pass

    @derived_from(algorithms.sequence_to_sequence.step)
    def step(self, k=1):
        pass

    @derived_from(algorithms.sequence_to_sequence.take)
    def take(self, number=1):
        pass

    @derived_from(algorithms.sequence_to_sequence.take_now, wrap_into_iterator=False)
    def take_now(self, number=1):
        pass

    def take_into_option(self, number=1):
        '''
        Returns chunk of *exactly* `number`-items from the beginning of iterator wrapped in `Some2`.
        In case of insufficient iterator returns `None2` and *returns* read items back to iterator.
        Note: If `None2` is returned then iterator is not invalidated!

        :param number:
        :return:

        Example:
        >>> it = Iterator2([1, 2, 3]); (it.take_into_option(4).is_none(), it.take_into_option(2).unwrap())
        (True, (1, 2))
        '''
        taken = self.take_now(number)
        if taken.length() == number:
            return Some2(taken)
        else:
            self._iterator = iter(taken)  # returns back, `taken` contains all items *before* take-call
            return None2

    @derived_from(algorithms.sequence_to_sequence.take_last)
    def take_last(self, number=1):
        pass

    @derived_from(algorithms.sequence_to_sequence.take_last, wrap_into_iterator=False)
    def take_last_now(self, number=1):
        pass

    def take_last_into_option(self, number=1):
        '''
        Returns chunk of *exactly* `number`-items from the end of iterator wrapped in `Some2`.
        In case of insufficient iterator returns `None2` and *returns* read items back to iterator.
        Note: If `None2` is returned then iterator is not invalidated!

        :param number:
        :return:

        Example:
        >>> it = Iterator2([1, 2, 3]); (it.take_into_option(4).is_none(), it.take_into_option(2).unwrap())
        (True, (1, 2))
        '''
        taken = self.take_last(number)
        if taken.length() == number:
            return Some2(taken)
        else:
            self._iterator = iter(taken)  # returns back, `taken` contains all items *before* take-call
            return None2

    @derived_from(algorithms.sequence_to_sequence.take_while)
    def take_while(self, predicate):
        pass

    # ===================
    # Checking conditions
    # ===================
    @derived_from(algorithms.predicate_on_sequence.all)
    def all(self, predicate):
        pass

    @derived_from(algorithms.predicate_on_sequence.any)
    def any(self, predicate):
        pass

    @derived_from(algorithms.predicate_on_sequence.none)
    def none(self, predicate):
        pass

    # =============
    # Making groups
    # =============
    @derived_from(algorithms.sequence_to_groups.chunks)
    def chunks(self, size, *, allow_partial=False):
        pass

    @derived_from(algorithms.sequence_to_groups.chunks_with_padding)
    def chunks_with_padding(self, size, *, fillvalue=None):
        pass

    @derived_from(algorithms.sequence_to_groups.consecutive_groups)
    def consecutive_groups(self, *, ordering=None):
        pass

    @based_on(algorithms.sequence_to_groups.group_by)
    def group_by(self, *, key=None):
        return Iterator2((
            (val, Iterator2(sub_iter))
            for val, sub_iter in self.raw()
        ))

    @derived_from(algorithms.sequence_to_groups.pairwise)
    def pairwise(self):
        pass

    @based_on(algorithms.sequence_to_groups.process_in_groups)
    def process_in_groups(self, *, key=None, transformation=None, aggregator=lambda it: Iterator2(it)):
        return Iterator2(self.raw())

    @derived_from(algorithms.sequence_to_groups.permutations)
    def permutations(self, repeat=None):
        pass

    @derived_from(algorithms.sequence_to_groups.sliding_window)
    def sliding_window(self, *, step=1, size=1, allow_partial=False):
        pass

    @derived_from(algorithms.sequence_to_groups.stagger)
    def stagger(self, *, offsets=(-1, 0, 1), fillvalue=None, longest=False):
        pass

    # ===========
    # Merging
    # ===========

    @derived_from(algorithms.merging.chain)
    def chain(self, *others):
        pass

    @derived_from(algorithms.merging.chain_from_iterable)
    def chain_from_iterable(self, others_in_iterable):
        pass

    @derived_from(algorithms.merging.interleave)
    def interleave(self, *others):
        pass

    @derived_from(algorithms.merging.interleave_shortest)
    def interleave_shortest(self, *others):
        pass

    @derived_from(algorithms.merging.prepend)
    def prepend(self, iterable):
        pass

    @derived_from(algorithms.merging.roundrobin)
    def roundrobin(self, *others):
        pass

    @derived_from(algorithms.merging.zip)
    def zip(self, *others):
        pass

    @derived_from(algorithms.merging.zip_longest)
    def zip_longest(self, *others, fillvalue=None):
        pass

    @derived_from(algorithms.merging.zip_offset)
    def zip_offset(self, *others, offsets, fillvalue=None, longest=False):
        pass

    # ===========
    # Splitting
    # ===========
    @based_on(algorithms.splitting.distribute)
    @with_rich_subiterators(in_tuple=True)
    def distribute(self, n):
        pass

    @based_on(algorithms.splitting.partition)
    @with_rich_subiterators(in_tuple=True)
    def partition(self, predicate):
        pass

    @based_on(algorithms.splitting.split_after)
    @with_rich_subiterators()
    def split_after(self, predicate):
        pass

    @based_on(algorithms.splitting.split_at)
    @with_rich_subiterators()
    def split_at(self, predicate):
        pass

    @based_on(algorithms.splitting.split_before)
    @with_rich_subiterators()
    def split_before(self, predicate):
        pass

    @based_on(algorithms.splitting.unzip)
    @with_rich_subiterators(in_tuple=True)
    def unzip(self, arity=2):
        pass

    # ===========
    # (Un)Merging
    # ===========
    @derived_from(algorithms.extra.cartesian_product)
    def cartesian_product(self, *others, repeat=1):
        pass

    # =================
    # Flow manipulation
    # =================
    def find(self, predicate):
        anti_predicate = lambda item: not predicate(item)
        return self.drop_while(anti_predicate).next()

    def first(self):
        return self.next()

    def last(self):
        return self.take_last_into_option(1)

    @derived_from(algorithms.searching.locate)
    def locate(self, predicate, *, count_from=0):
        pass

    def nth(self, index):
        return Iterator2(itertools_islice(self._iterator, index, index + 1)).next()

    @based_on(algorithms.searching.position)
    def position(self, predicate, *, count_from=0):
        idx = self._iterator
        return None2 if idx < 0 else Some2(idx)

    def sort(self, *, key=None, reverse=False):
        '''
        Sorts items from iterator with builtin `sorted`.

        :param key:
        :param reverse:
        :return:
        '''
        return Iterator2(
            sorted(self._iterator, key=key, reverse=reverse)
        )

    def unique(self):
        '''
        Removes equal items from iterator by means of `frozenset`. Order presuming is not guaranteed.

        :return:
        '''
        return Iterator2(frozenset(self.raw()))

    # =============
    # Materializing
    # =============
    def collect(self, factory=tuple):
        return factory(self.raw())

    def to_dict(self):
        return dict(self.raw())

    def to_frozenset(self):
        return frozenset(self.raw())

    def to_list(self):
        return list(self.raw())

    def to_set(self):
        return set(self.raw())

    def to_tuple(self):
        return tuple(self.raw())

    # ===========
    # Duplication
    # ===========
    def copy(self):
        self._iterator, it_copy = itertools_tee(self._iterator)  # implicit move semantics
        return Iterator2(it_copy)

    def tee(self, n=2):
        copies = itertools_tee(self.raw(), n)
        return tuple(map(Iterator2, copies))

    # =======
    # Folding
    # =======
    def apply(self, func):
        '''
        Uses iterator as an argument for `func.

        :param func:
        :return:
        '''
        return func(self.raw())

    @derived_from(algorithms.folding.count, wrap_into_iterator=False)
    def count(self):
        pass

    @derived_from(algorithms.folding.fold, wrap_into_iterator=False)
    def fold(self, func, *, initial=None):
        pass

    @derived_from(algorithms.folding.join, wrap_into_iterator=False)
    def join(self, sep):
        pass

    @derived_from(algorithms.folding.max, wrap_into_iterator=False)
    def max(self, *, default=None, key=None):
        pass

    @derived_from(algorithms.folding.min, wrap_into_iterator=False)
    def min(self, *, default=None, key=None):
        pass

    @derived_from(algorithms.folding.minmax, wrap_into_iterator=False)
    def minmax(self, *, default=None, key=None):
        pass

    @derived_from(algorithms.folding.product, wrap_into_iterator=False)
    def product(self):
        pass

    @derived_from(algorithms.folding.reduce, wrap_into_iterator=False)
    def reduce(self, func, *, initial=None):
        pass

    @derived_from(algorithms.folding.sum, wrap_into_iterator=False)
    def sum(self):
        pass

    # ======================
    # Lookahead and lookback
    # ======================
    def spy(self, n=1, *, allow_partial=False):
        if not isinstance(self._iterator, SpyIterator):
            self._iterator = SpyIterator(self._iterator)

        ok, items = self._iterator.spy(n, allow_partial=allow_partial)
        if ok:
            return Some2(items)
        else:
            return None2




import unittest

import iter2.algorithms.predicate_on_sequence as algo


class TestPredicateOnSequenceAlgorithms(unittest.TestCase):
    def test_all(self):
        with self.subTest('simple - positive'):
            self.assertEqual(
                algo.all([1, 2, 3], lambda x: x < 10),
                True
            )

        with self.subTest('simple - negative'):
            self.assertEqual(
                algo.all([1, 2, 3], lambda x: x % 2 == 0),
                False
            )

    def test_any(self):
        with self.subTest('simple - positive'):
            self.assertEqual(
                algo.any([1, 2, 3], lambda x: x % 2 == 0),
                True
            )

        with self.subTest('simple - negative'):
            self.assertEqual(
                algo.any([1, 2, 3], lambda x: x > 10),
                False
            )

    def test_none(self):
        with self.subTest('simple - positive'):
            self.assertEqual(
                algo.none([1, 2, 3], lambda x: x > 10),
                True
            )

        with self.subTest('simple - negative'):
            self.assertEqual(
                algo.none([1, 2, 3], lambda x: x % 2 == 0),
                False
            )





import unittest
import random

import iter2.algorithms.sequence_to_sequence as algo


class TestSequenceToSequenceAlgorithms(unittest.TestCase):
    def assertEqualAsTuples(self, first, second):
        self.assertEqual(
            tuple(first),
            tuple(second)
        )

    def test_accumulate(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.accumulate((1, 2, 3)),
                (1, 3, 6)
            )

        with self.subTest('custom func'):
            self.assertEqualAsTuples(
                algo.accumulate((1, 2, 3), lambda x, y: x * y),
                (1, 2, 6)
            )

    def test_add_side_effect(self):
        with self.subTest('simple'):
            side_effect_shell = []
            it = algo.add_side_effect(
                ['Julia', 'Angelina', 'Jessica'],
                side_effect_shell.append,
                before=lambda: side_effect_shell.append('before'),
                after=lambda: side_effect_shell.append('after')
            )
            res = tuple(it)
            self.assertEqualAsTuples(
                res,
                ('Julia', 'Angelina', 'Jessica')
            )
            self.assertEqualAsTuples(
                side_effect_shell,
                ('before', 'Julia', 'Angelina', 'Jessica', 'after')
            )

        with self.subTest('w/ chunks'):
            side_effect_shell = []
            it = algo.add_side_effect(
                range(10),
                side_effect_shell.append,
                chunk_size=4
            )
            res = tuple(it)
            self.assertEqualAsTuples(
                res,
                range(10)
            )
            self.assertEqualAsTuples(
                side_effect_shell,
                ((0, 1, 2, 3), (4, 5, 6, 7), (8, 9))
            )

    def test_consume(self):
        with self.subTest('full'):
            self.assertEqualAsTuples(
                algo.consume(range(10)),
                ()
            )

        with self.subTest('partial'):
            self.assertEqualAsTuples(
                algo.consume(range(10), 8),
                (8, 9)
            )

    def test_cycle(self):
        with self.subTest('fixed number of iterations'):
            self.assertEqualAsTuples(
                algo.cycle('ab', number=3),
                'ababab'
            )

        with self.subTest('infinite number of iterations'):
            self.assertTrue(
                all(
                    val == ans
                    for val, ans in zip(
                        algo.cycle('ab'),
                        'ab' * random.randint(1, 10)  # shame
                    )
                )
            )

    def test_dedup(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.dedup('aaaaabbbbbbbbc'),
                'abc'
            )

    def test_difference(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.difference((1, 3, 6)),
                (1, 2, 3)
            )

        with self.subTest('custom func'):
            self.assertEqualAsTuples(
                algo.difference((1, 2, 6), lambda x, y: x / y),
                (1, 2, 3)
            )

    def test_drop(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.drop('abcdef', 3),
                'def'
            )

    def test_drop_while(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.drop_while('abcDef', str.islower),
                'Def'
            )

    def test_enumerate(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.enumerate('abc'),
                ((0, 'a'), (1, 'b'), (2, 'c'))
            )

        with self.subTest('custom start index'):
            self.assertEqualAsTuples(
                algo.enumerate('abc', count_from=10),
                ((10, 'a'), (11, 'b'), (12, 'c'))
            )

    def test_filter(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.filter('aBcDe', str.islower),
                'ace'
            )

    def test_filter_none(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.filter_none((1, None, 2, None)),
                (1, 2)
            )

    def test_flatmap(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.flatmap(['one two', 'three four', 'five'], str.split),
                ('one', 'two', 'three', 'four', 'five')
            )

    def test_flatten(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.flatten(['ab', 'cd', 'ef']),
                'abcdef'
            )

    def test_intersperse(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.intersperse('abc', ' '),
                'a b c'
            )

    def test_map(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.map('abc', str.upper),
                'ABC'
            )

    def test_slice(self):
        with self.subTest('prefix'):
            self.assertEqualAsTuples(
                algo.slice('abcdef', 3),
                'abc'
            )

        with self.subTest('suffix'):
            self.assertEqualAsTuples(
                algo.slice('abcdef', 3, None),
                'def'
            )

        with self.subTest('step'):
            self.assertEqualAsTuples(
                algo.slice('abcdef', None, None, 2),
                'ace'
            )

        with self.subTest('complex'):
            self.assertEqualAsTuples(
                algo.slice('abcdef', 1, 5, 2),
                'bd'
            )

    def test_starmap(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.starmap(('aB', 'aC', 'zA'), lambda f, s: f.upper() + s.lower()),
                ('Ab', 'Ac', 'Za')
            )

    def test_step(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.step('abcdef', 3),
                'ad'
            )

    def test_take(self):
        with self.subTest('simple'):
            it = iter('abcdef')
            self.assertEqualAsTuples(
                algo.take(it, 3),
                'abc'
            )
            self.assertEqualAsTuples(
                it,
                'def'
            )

        with self.subTest('lazy'):
            it = iter('abcdef')
            take2 = algo.take(it, 2)
            take3 = algo.take(it, 3)
            self.assertEqualAsTuples(
                map(next, [take2, take3, take3]),
                'abc'
            )
            self.assertEqualAsTuples(take2, 'd')
            self.assertEqualAsTuples(take3, 'e')

        with self.subTest('partial'):
            self.assertEqualAsTuples(
                algo.take('ab', 3),
                'ab'
            )

    def test_take_now(self):
        with self.subTest('simple'):
            it = iter('abcdef')
            self.assertEqual(
                algo.take_now(it, 3),
                tuple('abc')
            )
            self.assertEqualAsTuples(it, 'def')

        with self.subTest('not lazy'):
            it = iter('abcdef')
            f3 = algo.take_now(it, 3)
            s2 = algo.take_now(it, 2)
            self.assertEqualAsTuples(f3, 'abc')
            self.assertEqualAsTuples(s2, 'de')
            self.assertEqualAsTuples(it, 'f')

        with self.subTest('partial'):
            self.assertEqual(
                algo.take_now('ab', 3),
                ('a', 'b')
            )

    def test_take_last(self):
        with self.subTest('simple & not lazy'):
            it = iter('abcdef')
            self.assertEqual(
                algo.take_last(it, 3),
                tuple('def')
            )
            self.assertEqualAsTuples(it, ())

        with self.subTest('partial'):
            self.assertEqual(
                algo.take_last('ab', 3),
                ('a', 'b')
            )

    def test_take_while(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.take_while('abcDef', str.islower),
                'abc'
            )



import unittest

from iter2.option import None2, Some2


class TestOption2(unittest.TestCase):
    def test_bool(self):
        with self.subTest('Some'):
            self.assertTrue(Some2(1))
        with self.subTest('None'):
            self.assertFalse(None2)

    def test_expect(self):
        with self.subTest('Some'):
            self.assertEqual(Some2(1).expect(RuntimeError), 1)
        with self.subTest('None'):
            with self.assertRaises(RuntimeError):
                None2.expect(RuntimeError, 'lol')

    def test_is_some(self):
        with self.subTest('Some'):
            self.assertTrue(Some2(1).is_some())
        with self.subTest('None'):
            self.assertFalse(None2.is_some())

    def test_is_none(self):
        with self.subTest('Some'):
            self.assertFalse(Some2(1).is_none())
        with self.subTest('None'):
            self.assertTrue(None2.is_none())

    def test_unwrap(self):
        with self.subTest('Some'):
            self.assertEqual(Some2(1).unwrap(), 1)
        with self.subTest('None'):
            with self.assertRaises(RuntimeError):
                None2.unwrap()

    def test_unwrap_or(self):
        with self.subTest('Some'):
            self.assertEqual(Some2(1).unwrap_or(2), 1)
        with self.subTest('None'):
            self.assertEqual(None2.unwrap_or(2), 2)

    def test_unwrap_or_else(self):
        with self.subTest('Some'):
            self.assertEqual(Some2(1).unwrap_or_else(lambda: 2), 1)
        with self.subTest('None'):
            self.assertEqual(None2.unwrap_or_else(lambda: 2), 2)

    def test_map(self):
        with self.subTest('Some'):
            self.assertEqual(Some2(-1).map(abs).unwrap(), 1)
        with self.subTest('None'):
            self.assertTrue(None2.map(abs).is_none())

    def test_map_or(self):
        with self.subTest('Some'):
            self.assertEqual(Some2(-1).map_or(2, abs), 1)
        with self.subTest('None'):
            self.assertEqual(None2.map_or(2, abs), 2)

    def test_map_or_else(self):
        with self.subTest('Some'):
            self.assertEqual(Some2(-1).map_or_else(lambda: 2, abs), 1)
        with self.subTest('None'):
            self.assertEqual(None2.map_or_else(lambda: 2, abs), 2)

    def test_and(self):
        with self.subTest('Some'):
            self.assertEqual(
                (Some2(object()) & Some2(1)).unwrap(),
                1
            )
        with self.subTest('None'):
            self.assertTrue(
                None2.and_(object()).is_none()
            )

    def test_and_then(self):
        with self.subTest('Some'):
            self.assertEqual(
                Some2(object()).and_then(lambda: Some2(1)).unwrap(),
                1
            )
        with self.subTest('None'):
            self.assertTrue(
                None2.and_then(lambda: Some2(object())).is_none()
            )

    def test_filter(self):
        with self.subTest('Some :: true'):
            self.assertEqual(Some2(1).filter(bool).unwrap(), 1)
        with self.subTest('Some :: false'):
            self.assertTrue(Some2(0).filter(bool).is_none())
        with self.subTest('None'):
            self.assertTrue(None2.filter(bool).is_none())

    def test_or(self):
        with self.subTest('Some'):
            self.assertEqual(
                Some2(1).or_(Some2(object())).unwrap(),
                1
            )
        with self.subTest('None'):
            self.assertEqual(None2 | 1, 1)

    def test_or_else(self):
        with self.subTest('Some'):
            self.assertEqual(
                Some2(1).or_else(lambda: Some2(object())).unwrap(),
                1
            )
        with self.subTest('None'):
            self.assertEqual(None2.or_else(lambda: 1), 1)

    def test_xor(self):
        with self.subTest('Some ^ Some'):
            self.assertTrue(
                Some2(1).xor(Some2(2)).is_none()
            )
        with self.subTest('Some ^ None'):
            self.assertEqual(
                Some2(1).xor(None2).unwrap(),
                1
            )
        with self.subTest('None ^ Some'):
            self.assertTrue(
                None2.xor(Some2(2)).unwrap(),
                2
            )
        with self.subTest('None ^ None'):
            self.assertTrue(
                None2.xor(None2).is_none()
            )



import unittest

import iter2.algorithms.sequence_to_groups as algo


class TestSequenceToGroupsAlgorithms(unittest.TestCase):
    def assertEqualAsTuples(self, first, second):
        self.assertEqual(
            tuple(first),
            tuple(second)
        )

    def test_chunks(self):
        with self.subTest('only complete'):
            self.assertEqualAsTuples(
                algo.chunks([1, 2, 3, 4], 2),
                ((1, 2), (3, 4))
            )

        with self.subTest('last incomplete'):
            self.assertEqualAsTuples(
                algo.chunks([1, 2, 3], 2),
                ((1, 2),)
            )

        with self.subTest('last incomplete but `allow_partial`'):
            self.assertEqualAsTuples(
                algo.chunks([1, 2, 3], 2, allow_partial=True),
                ((1, 2), (3,))
            )

    def test_chunks_with_padding(self):
        with self.subTest('only complete'):
            self.assertEqualAsTuples(
                algo.chunks_with_padding([1, 2, 3, 4], 2),
                ((1, 2), (3, 4))
            )

        with self.subTest('last incomplete'):
            self.assertEqualAsTuples(
                algo.chunks_with_padding([1, 2, 3], 2),
                ((1, 2), (3, None))
            )

        with self.subTest('last incomplete w/ custom `fillvalue`'):
            self.assertEqualAsTuples(
                algo.chunks_with_padding([1, 2, 3], 2, fillvalue=100500),
                ((1, 2), (3, 100500))
            )

    def test_consecutive_groups(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                map(tuple,
                    algo.consecutive_groups([1, 2, 10, 11, 100, 101])
                ),
                ((1, 2), (10, 11), (100, 101))
            )

        with self.subTest('custom `ordering`'):
            self.assertEqualAsTuples(
                map(tuple,
                    algo.consecutive_groups('abcBCDcde', ordering=ord)
                ),
                (('a', 'b', 'c'), ('B', 'C', 'D'), ('c', 'd', 'e'))
            )

    def test_group_by(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                (
                    (key, tuple(group))
                    for key, group in algo.group_by([1, 2, 2, 3, 3, 3])
                ),
                ((1, (1,)), (2, (2, 2)), (3, (3, 3, 3)))
            )

        with self.subTest('w/ key'):
            self.assertEqualAsTuples(
                (
                    (key, tuple(group))
                    for key, group in algo.group_by([1, 2, 2, 3, 3, 3])
                ),
                ((1, (1,)), (2, (2, 2)), (3, (3, 3, 3)))
            )

    def test_pairwise(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.pairwise([1, 2, 3]),
                ((1, 2), (2, 3))
            )

    def test_permutations(self):
        with self.subTest('n=2'):
            self.assertEqualAsTuples(
                algo.permutations([0, 1], length=2),
                ((0, 1), (1, 0))
            )

        with self.subTest('n=2'):
            self.assertEqualAsTuples(
                algo.permutations([0, 1, 2], length=2),
                ((0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1))
            )

    def test_process_in_groups(self):
        materialize = lambda it: tuple((key, tuple(val)) for key, val in it)

        with self.subTest('"group_by"'):
            self.assertEqual(
                materialize(
                    algo.process_in_groups([1, 2, 2, 3, 3, 3], key=lambda x: x % 2)
                ),
                ((1, (1,)), (0, (2, 2)), (1, (3, 3, 3)))
            )

        with self.subTest('w/ transformation'):
            self.assertEqual(
                materialize(
                    algo.process_in_groups([1, 2, 2, 3, 3, 3], transformation=lambda x: x % 2)
                ),
                ((1, (1,)), (2, (0, 0)), (3, (1, 1, 1)))
            )

        with self.subTest('w/ aggregator'):
            self.assertEqualAsTuples(
                algo.process_in_groups([1, 2, 2, 3, 3, 3], aggregator=sum),
                ((1, 1), (2, 4), (3, 9))
            )

    def test_sliding_window(self):
        with self.subTest('"chunks"; `allow_partial` is False"'):
            self.assertEqualAsTuples(
                map(tuple,
                    algo.sliding_window([1, 2, 3, 4, 5], size=2, step=2),
                ),
                ((1, 2), (3, 4))
            )

        with self.subTest('"chunks"; `allow_partial` is True'):
            self.assertEqualAsTuples(
                map(tuple,
                    algo.sliding_window([1, 2, 3, 4, 5], size=2, step=2, allow_partial=True),
                ),
                ((1, 2), (3, 4), (5,))
            )

        with self.subTest('overlapping'):
            self.assertEqualAsTuples(
                map(tuple,
                    algo.sliding_window([1, 2, 3, 4, 5], size=2, step=1),
                ),
                ((1, 2), (2, 3), (3, 4), (4, 5))
            )

        with self.subTest('wide step'):
            self.assertEqualAsTuples(
                map(tuple,
                    algo.sliding_window([1, 2, 3, 4, 5], size=2, step=3),
                    ),
                ((1, 2), (4, 5))
            )

    def test_stagger(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.stagger([0, 1, 2, 3]),
                ((None, 0, 1), (0, 1, 2), (1, 2, 3))
            )

        with self.subTest('custom `offsets`'):
           self.assertEqualAsTuples(
                algo.stagger([0, 1, 2, 3, 4, 5, 6, 7], offsets=(0, 2, 4)),
                ((0, 2, 4), (1, 3, 5), (2, 4, 6), (3, 5, 7))
           )

        with self.subTest('w/ `longest`'):
            self.assertEqualAsTuples(
                algo.stagger([0, 1, 2, 3], longest=True),
                ((None, 0, 1), (0, 1, 2), (1, 2, 3), (2, 3, None), (3, None, None))
            )



import unittest

import iter2.algorithms.searching as algo


class TestSearchingAlgorithms(unittest.TestCase):
    def test_find(self):
        with self.subTest('found'):
            self.assertEqual(
                algo.find('aBc', lambda c: c.isupper()),
                'B'
            )

        with self.subTest('not found'):
            with self.assertRaises(StopIteration):
                algo.find('abc', lambda c: c.isupper())


    def test_locate(self):
        with self.subTest('w/o items'):
            self.assertEqual(
                tuple(algo.locate('aBcDe', str.isupper, count_from=1)),
                (2, 4)
            )

        with self.subTest('w/ items'):
            self.assertEqual(
                tuple(algo.locate('aBcDe', str.isupper, count_from=1, with_items=True)),
                ((2, 'B'), (4, 'D'))
            )

        with self.subTest('not found'):
            self.assertEqual(
                tuple(algo.locate('abcde', str.isupper, count_from=1)),
                ()
            )

    def test_position(self):
        with self.subTest('found'):
            self.assertEqual(
                algo.position('aBcDe', str.isupper, count_from=1),
                2
            )

        with self.subTest('not found'):
            self.assertEqual(
                algo.position('abcde', str.isupper, count_from=1),
                -1
            )



import setuptools


with open("README.md", "r") as readme_file:
    long_description = readme_file.read()


setuptools.setup(
    name='iter2',
    version='1.0rc',
    description='Implementation of some rich-iterator concept',
    author='Shemyakin Roman',
    author_email='ramon93i7@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ramon93i7/iter2',
    license='MIT',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License'
    ]
)



from . import sequence_to_sequence
from . import sequence_to_groups
from . import merging
from . import splitting
from . import searching
from . import folding
from . import predicate_on_sequence
from . import extra



import unittest

import iter2.algorithms.merging as algo


class TestMergingAlgorithms(unittest.TestCase):
    def assertEqualAsTuples(self, first, second):
        self.assertEqual(
            tuple(first),
            tuple(second)
        )

    def test_chain(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.chain([1, 2], [3, 4]),
                (1, 2, 3, 4)
            )

    def test_chain_from_iterable(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.chain_from_iterable([[1, 2], [3, 4]]),
                (1, 2, 3, 4)
            )

    def test_collate(self):
        pass

    def test_interleave(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.interleave('123', 'abc'),
                ('1', 'a', '2', 'b', '3', 'c')
            )

        with self.subTest('asymmetric'):
            self.assertEqualAsTuples(
                algo.interleave('12', 'abc'),
                ('1', 'a', '2', 'b', 'c')
            )

    def test_interleave_shortest(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.interleave_shortest('123', 'abc'),
                ('1', 'a', '2', 'b', '3', 'c')
            )

        with self.subTest('asymmetric'):
            self.assertEqualAsTuples(
                algo.interleave_shortest('12', 'abc'),
                ('1', 'a', '2', 'b')
            )

    def test_prepend(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.prepend([3, 4], [1, 2]),
                (1, 2, 3, 4)
            )

    def test_zip(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.zip('123', 'abc'),
                (('1', 'a'), ('2', 'b'), ('3', 'c'))
            )

        with self.subTest('asymmetric'):
            self.assertEqualAsTuples(
                algo.zip('12', 'abc'),
                (('1', 'a'), ('2', 'b'))
            )

    def test_zip_longest(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.zip_longest('123', 'abc'),
                (('1', 'a'), ('2', 'b'), ('3', 'c'))
            )

        with self.subTest('asymmetric'):
            self.assertEqualAsTuples(
                algo.zip_longest('12', 'abc', fillvalue='#'),
                (('1', 'a'), ('2', 'b'), ('#', 'c'))
            )

    def test_zip_offset(self):
        with self.subTest('simple'):
            self.assertEqualAsTuples(
                algo.zip_offset('12', 'abcd', offsets=(0, 1)),
                (('1', 'b'), ('2', 'c'))
            )

        with self.subTest('longest'):
            self.assertEqualAsTuples(
                algo.zip_offset('12', 'abcd', offsets=(0, 1), fillvalue='#', longest=True),
                (('1', 'b'), ('2', 'c'), ('#', 'd'))
            )



import unittest

import iter2.algorithms.folding as algo


class TestFoldingAlgorithms(unittest.TestCase):
    def test_count(self):
        with self.subTest('simple'):
            self.assertEqual(
                algo.count([1, 2, 3]),
                3
            )

    def test_fold(self):
        with self.subTest('simple'):
            self.assertEqual(
                algo.fold([1, 2, 5, 4, 3], max),
                5
            )

        with self.subTest('w/ initial'):
            self.assertEqual(
                algo.fold([1, 2, 3, 4, 5], max, initial=100500),
                100500
            )

    def test_join(self):
        with self.subTest('simple'):
            self.assertEqual(
                algo.join(['1', '2', '3'], '-'),
                '1-2-3'
            )

    def test_max(self):
        with self.subTest('simple'):
            self.assertEqual(
                algo.max([1, 2, 3, 4, 5]),
                5
            )

        with self.subTest('w/ default and no items'):
            self.assertEqual(
                algo.max([], default=100500),
                100500
            )

        with self.subTest('w/ default and some items'):
            self.assertEqual(
                algo.max([1, 2], default=100500),
                2
            )

        with self.subTest('w/ key function'):
            self.assertEqual(
                algo.max(
                    [None, object()],
                    key=lambda item: 100500 if item is None else 0
                ),
                None
            )

    def test_min(self):
        with self.subTest('simple'):
            self.assertEqual(
                algo.min([1, 2, 3, 4, 5]),
                1
            )

        with self.subTest('w/ default and no items'):
            self.assertEqual(
                algo.min([], default=100500),
                100500
            )

        with self.subTest('w/ default and some items'):
            self.assertEqual(
                algo.min([1, 2], default=100500),
                1
            )

        with self.subTest('w/ key function'):
            self.assertEqual(
                algo.min(
                    [None, object()],
                    key=lambda item: 0 if item is None else 100500
                ),
                None
            )

    def test_min_max(self):
        with self.subTest('simple'):
            self.assertEqual(
                algo.minmax([1, 2, 3, 4, 5]),
                (1, 5)
            )

        with self.subTest('w/ default and no items'):
            self.assertEqual(
                algo.minmax([], default=100500),
                100500
            )

        with self.subTest('w/ default and some items'):
            self.assertEqual(
                algo.minmax([1, 2], default=100500),
                (1, 2)
            )

        with self.subTest('w/ key function'):
            self.assertEqual(
                algo.minmax(
                    (True, None),
                    key=lambda item: 0 if item is None else 100500
                ),
                (None, True)
            )

    def test_product(self):
        with self.subTest('simple'):
            self.assertEqual(
                algo.product([1, 2, 3, 4, 5]),
                120
            )

    def test_sum(self):
        with self.subTest('simple'):
            self.assertEqual(
                algo.sum([1, 2, 3, 4, 5]),
                15
            )



import itertools
import operator
import functools

from .iterators.iterator2 import Iterator2

from .utils import define_module_exporter


export_from_module, __all__ = define_module_exporter()  # setup exports


# Aliases
builtin_range = range
builtin_map = map

itertools_chain = itertools.chain
itertools_count = itertools.count
itertools_cycle = itertools.cycle
itertools_repeat = itertools.repeat
itertools_takewhile = itertools.takewhile

operator_gt = operator.gt
operator_lt = operator.lt
functools_partial = functools.partial


@export_from_module
def count_from(start=0, *, step=1):
    return Iterator2(itertools_count(start, step))


@export_from_module
def cycle(iterable):
    return Iterator2(itertools_cycle(iterable))


@export_from_module
def empty():
    return Iterator2(())


@export_from_module
def from_args(*items):
    return Iterator2(items)


@export_from_module
def iterate(obj, func):
    return Iterator2(itertools_chain((obj,), itertools_repeat(func))).accumulate(lambda st, f: f(st))


@export_from_module
def numeric_range(*args):
    argc = len(args)
    if argc == 1:
        stop, = args
        start = type(stop)(0)
        step = 1
    elif argc == 2:
        start, stop = args
        step = 1
    elif argc == 3:
        start, stop, step = args
    else:
        err_msg = 'numeric_range takes at most 3 arguments, got {}'
        raise TypeError(err_msg.format(argc))

    values = (start + (step * n) for n in itertools_count())
    if step > 0:
        return Iterator2(itertools_takewhile(functools_partial(operator_gt, stop), values))
    elif step < 0:
        return Iterator2(itertools_takewhile(functools_partial(operator_lt, stop), values))
    else:
        raise ValueError('numeric_range arg 3 (`step`) must not be zero')


@export_from_module
def once(obj):
    return Iterator2((obj,))


@export_from_module
def range(*args, **kwargs):
    return Iterator2(builtin_range(*args, **kwargs))


@export_from_module
def repeat(object, times=None):
    if times is None:
        return Iterator2(itertools_repeat(object))
    else:
        return Iterator2(itertools_repeat(object, times))


@export_from_module
def repeat_call(func, times=None):
    return repeat(func, times=times).map(lambda f: f())


@export_from_module
def tabulate(func, start=0):
    return Iterator2(builtin_map(func, itertools_count(start)))
